/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::application::AronArmemGenerator
 * @author     fabian.peller-konrad@kit.edu ( fabian dot peller-konrad at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// STD/STL
#include <iostream>
#include <filesystem>
#include <ctime>

// Boost
#include <boost/algorithm/string.hpp>

// CXXOPTS
#include <RobotAPI/applications/AronGenerator/cxxopts.hpp>

// ArmarX Executable
//#include <ArmarXCore/core/application/Application.h>
//#include <ArmarXCore/core/Component.h>
//#include <ArmarXCore/core/logging/Logging.h>

// ArmarX
#include <ArmarXCore/libraries/cppgen/CppMethod.h>
#include <ArmarXCore/libraries/cppgen/CppClass.h>

#include <RobotAPI/libraries/aron/codegenerator/typeReaders/xmlReader/AronAbstractTypeXMLReader.h>
#include <RobotAPI/libraries/aron/codegenerator/codeWriters/cppWriter/AronCppWriter.h>

using namespace armarx;
using namespace aron;

int main(int argc, char* argv[])
{
    try
    {
        cxxopts::Options options("AronArmemCodeGenerator", "An application interface for the aron and armem code generation");

        options.add_options("General")
        ("v,verbose", "Enable verbose mode", cxxopts::value<bool>()->default_value("false"))
        ("h,help", "Print usage");

        options.add_options("file")
        ("f,file", "XML file name", cxxopts::value<std::string>()->default_value("/path/to/some/xml/file.xml"))
        ("o,output", "The output path", cxxopts::value<std::string>()->default_value("/path/to/some/output/folder/"));

        auto result = options.parse(argc, argv);

        if (result.count("help"))
        {
            std::cout << options.help() << std::endl;
            exit(0);
        }

        bool verbose = result["v"].as<bool>();
        std::string filename = result["f"].as<std::string>();
        std::string output = result["o"].as<std::string>();

        codegeneration::typereader::AronAbstractTypeXMLReader reader;

        if (verbose)
        {
            std::cout << "Welcome to the AronGenerator!" << std::endl;
            std::cout << "Received the following parameters:" << std::endl;
            std::cout << "\tFile: " << filename << std::endl;
            std::cout << "\tOutput: " << output << std::endl;
            std::cout << std::endl;
        }

        std::cout << "Generating a new cpp file out of <" + filename + ">" << std::endl;

        if (verbose)
        {
            std::cout << "Parsing the XML file...";
        }

        std::filesystem::path input_file(filename);
        if (!std::filesystem::exists(input_file))
        {
            std::cerr << "The input file does not exist!" << std::endl;
        }

        if (input_file.extension().string() != ".xml")
        {
            std::cerr << "The file you passed has the wrong type." << std::endl;
        }

        std::string new_name_with_extension = input_file.filename().replace_extension("").string();
        new_name_with_extension += ".aron.generated.h";
        std::string output_file_path = output + "/" + new_name_with_extension;
        std::filesystem::path output_file(output_file_path);

        reader.parseFile(input_file);
        if (verbose)
        {
            std::cout << " done!" << std::endl;
            std::cout << "--> Found " << reader.getGenerateTypes().size() << " types." << std::endl;
        }

        codegeneration::classwriter::AronTypeClassCppWriter writer("AronTestSegment", reader.getWriters(), reader.getReaders(), reader.getUseTypes(), reader.getAdditionalIncludes());

        if (verbose)
        {
            std::cout << "Running the type class generator...";
        }

        writer.generateTypeClasses(reader.getGenerateTypes());

        if (verbose)
        {
            std::cout << " done!" << std::endl;
            std::cout << "--> Found " << writer.getTypeClasses().size() << " type classes." << std::endl;
        }

        std::stringstream file_new_content;

        file_new_content << "\
/* \n\
 * This file is part of ArmarX. \n\
 * \n\
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T), \n\
 * Karlsruhe Institute of Technology (KIT), all rights reserved. \n\
 * \n\
 * ArmarX is free software; you can redistribute it and/or modify \n\
 * it under the terms of the GNU General Public License version 2 as \n\
 * published by the Free Software Foundation. \n\
 * \n\
 * ArmarX is distributed in the hope that it will be useful, but \n\
 * WITHOUT ANY WARRANTY; without even the implied warranty of \n\
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the \n\
 * GNU General Public License for more details. \n\
 * \n\
 * You should have received a copy of the GNU General Public License \n\
 * along with this program. If not, see <http://www.gnu.org/licenses/>. \n\
 * \n\
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt \n\
 *             GNU General Public License \n\
 ************************************************************************* \n\
 * WARNING: This file is autogenerated. \n\
 * Original file: " + filename + " \n\
 * Please do not edit since your changes may be overwritten on the next generation \n\
 * If you have any questions please contact: Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu) \n\
 ************************************************************************* \n\
 */\n\n\n";

        unsigned int i = 0;
        for (const auto& c : writer.getTypeClasses())
        {
            i++;
            if (verbose)
            {
                std::cout << "Exporting class '" << c->getName() << "'...";
            }

            CppWriterPtr w = CppWriterPtr(new CppWriter());
            c->writeCpp(w);

            file_new_content << w->getString();

            if (verbose)
            {
                std::cout << " done!" << std::endl;
            }

            if (i != writer.getTypeClasses().size())
            {
                file_new_content << "\n";
                file_new_content << "// *************\n";
                file_new_content << "// Here comes another auto-generated class. Don't mind the duplicated includes - everything is guarded!\n";
                file_new_content << "// *************\n";
                file_new_content << "\n";
            }
        }

        // check if file already exists
        if (std::filesystem::exists(output_file))
        {
            std::ifstream ifs(output_file_path);
            std::string file_content((std::istreambuf_iterator<char>(ifs)), (std::istreambuf_iterator<char>()));

            if (file_content == file_new_content.str())
            {
                std::cout << "File content not changed for <" + output_file_path + ">" << std::endl;
                exit(0);
            }
        }

        std::ofstream ofs;
        ofs.open(output_file_path);
        ofs << file_new_content.str();
        ofs.close();

        std::cout << "Finished generating <" + output_file_path + ">" << std::endl;
    }
    catch (const cxxopts::OptionException& e)
    {
        std::cerr << "Error in parsing cxxopts options: " << e.what() << std::endl;
        exit(1);
    }
    exit(0);
}
