#pragma once

#include <type_traits>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/interface/ArViz/Elements.h>

#include "ElementOps.h"
#include "point_cloud_type_traits.hpp"


namespace armarx::viz
{

    using data::ColoredPoint;


    class PointCloud : public ElementOps<PointCloud, data::ElementPointCloud>
    {
    public:

        using ElementOps::ElementOps;

        PointCloud& clear()
        {
            data_->points.clear();
            return *this;
        }

        PointCloud& transparency(float t)
        {
            data_->transparency = t;

            return *this;
        }

        PointCloud& points(std::vector<ColoredPoint> const& ps)
        {
            data_->points = ps;

            return *this;
        }

        PointCloud& addPoint(ColoredPoint const& p)
        {
            data_->points.push_back(p);

            return *this;
        }

        PointCloud& addPoint(float x, float y, float z, const data::Color& color)
        {
            ColoredPoint p;
            p.x = x;
            p.y = y;
            p.z = z;
            p.color = color;
            return addPoint(p);
        }

        PointCloud& addPoint(float x, float y, float z, const simox::Color& color)
        {
            return addPoint(x, y, z, Color{color});
        }

        template <typename ColorCoeff = int>
        PointCloud & addPoint(float x, float y, float z, ColorCoeff r, ColorCoeff g, ColorCoeff b, ColorCoeff a = 255)
        {
            return addPoint(x, y, z, simox::Color(r, g, b, a));
        }

        PointCloud& addPoint(float x, float y, float z)
        {
            return addPoint(x, y, z, simox::Color::black(255));
        }

        PointCloud& addPoint(float x, float y, float z, std::size_t id, int alpha = 255)
        {
            return addPoint(x, y, z, simox::color::GlasbeyLUT::at(id, alpha));
        }

        // Templated setters for PCL point clouds.

        template < class PointT,
                   std::enable_if_t < !detail::has_members_rgba<PointT>::value, int > = 0 >
        PointCloud & addPoint(const PointT& p, Color color = Color::gray())
        {
            return addPoint(p.x, p.y, p.z, color);
        }

        template < class PointT,
                   std::enable_if_t < detail::has_members_rgba<PointT>::value, int> = 0>
        PointCloud & addPoint(const PointT& p)
        {
            return addPoint(p.x, p.y, p.z, simox::Color(p.r, p.g, p.b, p.a));
        }

        template < class PointT >
        PointCloud& addPoint(const PointT& p, Color color)
        {
            return addPoint(p.x, p.y, p.z, color);
        }


        /// Draw a point cloud.
        template <class PointCloudT>
        PointCloud& pointCloud(const PointCloudT& cloud, bool checkFinite = true)
        {
            clear();
            for (const auto& p : cloud)
            {
                if (isfinite(p, checkFinite))
                {
                    addPoint(p);
                }
            }
            return *this;
        }

        /// Draw a point cloud with fixed color.
        template <class PointCloudT>
        PointCloud& pointCloud(const PointCloudT& cloud, Color color, bool checkFinite = true)
        {
            clear();
            for (const auto& p : cloud)
            {
                if (isfinite(p, checkFinite))
                {
                    addPoint(p, color);
                }
            }
            return *this;
        }

        /// Draw a point cloud with given indices.
        template <class PointCloudT>
        PointCloud& pointCloud(const PointCloudT& cloud, const std::vector<int>& indices,
                               bool checkFinite = true)
        {
            clear();
            for (int i : indices)
            {
                ARMARX_CHECK_FITS_SIZE(i, cloud.size());
                const auto& p = cloud.at(size_t(i));
                if (isfinite(p, checkFinite))
                {
                    addPoint(p);
                }
            }
            return *this;
        }

        /// Draw a point cloud with given indices and fixed color.
        template <class PointCloudT>
        PointCloud& pointCloud(const PointCloudT& cloud, const std::vector<int>& indices,
                               Color color, bool checkFinite = true)
        {
            clear();
            for (int i : indices)
            {
                ARMARX_CHECK_FITS_SIZE(i, cloud.size());
                const auto& p = cloud.at(size_t(i));
                if (isfinite(p, checkFinite))
                {
                    addPoint(p, color);
                }
            }
            return *this;
        }


        PointCloud& pointSizeInPixels(float s)
        {
            data_->pointSizeInPixels = s;
            return *this;
        }


    private:

        template <class PointT>
        static bool isfinite(const PointT& p, bool checkFinite = true)
        {
            return !checkFinite || (std::isfinite(p.x) && std::isfinite(p.y) && std::isfinite(p.z));
        }

    };

}


