/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


// Boost
#include <boost/algorithm/string.hpp>

// Header
#include "AronCppWriter.h"

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>

namespace armarx
{
    namespace aron
    {
        namespace codegeneration
        {
            namespace classwriter
            {
                AronTypeClassCppWriter::AronTypeClassCppWriter(const std::string& producerName, const std::set<AronWriterInfo>& writers, const std::set<AronReaderInfo>& readers, const std::vector<AronUseTypeInfo>& useTypes, const std::vector<std::string>& additionalIncludesFromFile) :
                    producerName(producerName),
                    writers(writers),
                    readers(readers),
                    additionalIncludes(additionalIncludesFromFile)
                {
                    additionalIncludes.push_back("<RobotAPI/libraries/aron/codegenerator/AronCppClass.h>");
                    for (const auto& useType : useTypes)
                    {
                        additionalIncludes.push_back(useType.include);
                    }
                }

                std::vector<CppClassPtr> AronTypeClassCppWriter::getTypeClasses() const
                {
                    return typeClasses;
                }

                void AronTypeClassCppWriter::generateTypeClasses(const GenerateTypePtrList& produceTypes, bool onlyAddHeaderInformationToFirstClass)
                {
                    bool first = true;
                    for (const auto& [produceObj, localProduceTypes] : produceTypes)
                    {
                        // Convert to Object type and create class object
                        if (produceObj == nullptr)
                        {
                            throw LocalException("An received produce type is not an object. You can only produce objects!");
                        }

                        CppClassPtr c = setupCppClassFromObjectType(produceObj);
                        c->addInherit("virtual public armarx::aron::codegeneration::AronCppClass");

                        // Add includes and guard
                        if (!onlyAddHeaderInformationToFirstClass or first)
                        {
                            setupIncludes(c, boost::algorithm::to_upper_copy(producerName) + "__" + boost::algorithm::to_upper_copy(c->getName()) + "__ARON_TYPE_DEFINITION_INCLUDE_GUARD");
                        }

                        // setup inner classes
                        for (const auto& localProduceObj : localProduceTypes)
                        {
                            CppClassPtr inner = setupCppClassFromObjectType(localProduceObj);
                            CppMethodPtr inner_equals = localProduceObj->toEqualsMethod();
                            inner->addMethod(inner_equals);

                            c->addInnerClass(inner);
                        }

                        // ctor
                        CppBlockPtr ctorBlock = CppBlockPtr(new CppBlock());
                        ctorBlock->addLine("initialize();");
                        c->addCtor("", ctorBlock);

                        // Generic methods
                        CppMethodPtr equals = produceObj->toEqualsMethod();
                        c->addMethod(equals);

                        CppMethodPtr reset = produceObj->toResetMethod();
                        c->addMethod(reset);

                        CppMethodPtr setup = produceObj->toInitializeMethod();
                        c->addMethod(setup);

                        CppMethodPtr write = produceObj->toWriteMethod();
                        c->addMethod(write);

                        CppMethodPtr read = produceObj->toReadMethod();
                        c->addMethod(read);

                        // Writermethods
                        for (const AronWriterInfo& info : writers)
                        {
                            c->addInclude(info.include);
                            CppMethodPtr convert = produceObj->toSpecializedWriterMethod(info.returnType, info.methodName, info.writerClassType);
                            c->addMethod(convert);
                        }

                        // Add methods to set the member variables
                        for (const AronReaderInfo& info : readers)
                        {
                            c->addInclude(info.include);
                            CppMethodPtr convert = produceObj->toSpecializedReaderMethod(info.argumentType, info.methodName, info.readerClassType);
                            c->addMethod(convert);
                        }

                        typeClasses.push_back(c);
                        first = false;
                    }
                }

                CppClassPtr AronTypeClassCppWriter::setupCppClassFromObjectType(const types::AronObjectTypeNavigatorPtr& o) const
                {
                    std::vector<std::string> split;
                    std::string cppTypename = o->getFullCppTypename();
                    boost::algorithm::split(split, cppTypename, boost::is_any_of("::"), boost::token_compress_on);
                    if (split.size() < 1)
                    {
                        throw LocalException("AronTypeClassCppWriter: The cpp name of an inner type was empty. Please check the type definition file if all object names are set correctly.");
                    }

                    std::vector<std::string> namespaces(split);
                    namespaces.pop_back();

                    CppClassPtr c = CppClassPtr(new CppClass(namespaces, split[split.size() - 1]));

                    for (const auto& [key, child] : o->getChildren())
                    {
                        c->addPublicField(child->getFullCppTypename() + " " + key + ";");
                    }

                    return c;
                }

                void AronTypeClassCppWriter::setupIncludes(CppClassPtr& c, const std::string& guard) const
                {
                    c->addClassDoc("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
                                   "!!!!!!AUTOGENERATED CLASS. Please do NOT edit. Instead, edit the corresponding .xml file!!!!!!\n"
                                   "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                    c->setIncludeGuard(guard);
                    c->addInclude("<string>");
                    c->addInclude("<vector>");
                    c->addInclude("<map>");
                    c->addInclude("<RobotAPI/interface/aron.h>");
                    for (const std::string& s : additionalIncludes)
                    {
                        c->addInclude(s);
                    }
                }
            }
        }
    }
}





