/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// STD/STL


// Boost
#include <boost/algorithm/string.hpp>

// Header
#include "AronAbstractTypeXMLData.h"

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>
#include <ArmarXCore/core/rapidxml/wrapper/RapidXmlReader.h>

namespace armarx
{
    namespace aron
    {
        namespace codegeneration
        {
            namespace typereader
            {
                const std::string AronAbstractTypeXMLReaderData::ARON_TYPE_DEFINITION_TAG = "arontypedefinition";

                const std::string AronAbstractTypeXMLReaderData::ARON_INCLUDES_TAG = "includes";
                const std::string AronAbstractTypeXMLReaderData::ARON_SERIALIZERS_TAG = "serializers";
                const std::string AronAbstractTypeXMLReaderData::ARON_USE_TYPES_TAG = "usetypes";
                const std::string AronAbstractTypeXMLReaderData::ARON_PRODUCE_TYPES_TAG = "generatetypes";

                const std::string AronAbstractTypeXMLReaderData::ARON_INCLUDE_TAG = "include";
                const std::string AronAbstractTypeXMLReaderData::ARON_WRITERS_TAG = "writers";
                const std::string AronAbstractTypeXMLReaderData::ARON_READERS_TAG = "readers";

                const std::string AronAbstractTypeXMLReaderData::ARON_WRITER_TAG = "writer";
                const std::string AronAbstractTypeXMLReaderData::ARON_READER_TAG = "reader";
                const std::string AronAbstractTypeXMLReaderData::ARON_USE_TYPE_TAG = "usetype";
                const std::string AronAbstractTypeXMLReaderData::ARON_PRODUCE_TYPE_TAG = "generatetype";

                const std::string AronAbstractTypeXMLReaderData::ARON_INCLUDE_ATTRIBUTE_NAME = "include";
                const std::string AronAbstractTypeXMLReaderData::ARON_METHOD_ATTRIBUTE_NAME = "method";
                const std::string AronAbstractTypeXMLReaderData::ARON_RETURN_ATTRIBUTE_NAME = "return";
                const std::string AronAbstractTypeXMLReaderData::ARON_ARGUMENT_TYPE_ATTRIBUTE_NAME = "argumenttype";
                const std::string AronAbstractTypeXMLReaderData::ARON_WRITER_ATTRIBUTE_NAME = "writer";
                const std::string AronAbstractTypeXMLReaderData::ARON_READER_ATTRIBUTE_NAME = "reader";
                const std::string AronAbstractTypeXMLReaderData::ARON_NAME_ATTRIBUTE_NAME = "name";
                const std::string AronAbstractTypeXMLReaderData::ARON_TYPE_ATTRIBUTE_NAME = "type";
                const std::string AronAbstractTypeXMLReaderData::ARON_KEY_ATTRIBUTE_NAME = "key";
                const std::string AronAbstractTypeXMLReaderData::ARON_WIDTH_ATTRIBUTE_NAME = "width";
                const std::string AronAbstractTypeXMLReaderData::ARON_HEIGHT_ATTRIBUTE_NAME = "height";

                const std::string AronAbstractTypeXMLReaderData::ARON_OBJECT_CHILD_TAG = "objectchild";
                const std::string AronAbstractTypeXMLReaderData::ARON_TYPE_TAG = "type";
                const std::string AronAbstractTypeXMLReaderData::ARON_DIMENSIONS_TAG = "dimensions";
                const std::string AronAbstractTypeXMLReaderData::ARON_DIMENSION_TAG = "dimension";

                // Top Level type tags
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    const std::string AronAbstractTypeXMLReaderData::ARON_PRODUCE_##capsType##_TAG = #lowerType;

                HANDLE_ALL_ARON_TYPES_WITHOUT_CPP_TYPE
#undef RUN_ARON_MACRO

                void AronAbstractTypeXMLReaderData::EnforceAttribute(const RapidXmlReaderNode& node, const std::string& att)
                {
                    if (!AronAbstractTypeXMLReaderData::HasAttribute(node, att))
                    {
                        throw LocalException("AronAbstractTypeXMLReaderData: A <" + node.name() + ">-tag does not have the correct attribute. Expected: " + att);
                    }
                }

                bool AronAbstractTypeXMLReaderData::HasAttribute(const RapidXmlReaderNode& node, const std::string& att)
                {
                    return node.has_attribute(att.c_str());
                }

                bool AronAbstractTypeXMLReaderData::HasTagName(const RapidXmlReaderNode& node, const std::string& name)
                {
                    return (boost::algorithm::to_lower_copy(name) == boost::algorithm::to_lower_copy(node.name()));
                }

                void AronAbstractTypeXMLReaderData::EnforceTagName(const RapidXmlReaderNode& node, const std::string& name)
                {
                    if (!(HasTagName(node, name)))
                    {
                        throw LocalException("AronAbstractTypeXMLReaderData: The node <" + node.name() + "> has the wrong tag. Expected: <" + name + ">");
                    }
                }

                bool AronAbstractTypeXMLReaderData::CheckMinChildSize(const RapidXmlReaderNode& node, const size_t size)
                {
                    std::vector<RapidXmlReaderNode> children = node.nodes();
                    return children.size() >= size;
                }

                bool AronAbstractTypeXMLReaderData::CheckMaxChildSize(const RapidXmlReaderNode& node, const size_t size)
                {
                    std::vector<RapidXmlReaderNode> children = node.nodes();
                    return children.size() <= size;
                }

                bool AronAbstractTypeXMLReaderData::CheckExactChildSize(const RapidXmlReaderNode& node, const size_t size)
                {
                    std::vector<RapidXmlReaderNode> children = node.nodes();
                    return children.size() == size;
                }

                void AronAbstractTypeXMLReaderData::EnforceChildSize(const RapidXmlReaderNode& node, const size_t size)
                {
                    if (!AronAbstractTypeXMLReaderData::CheckExactChildSize(node, size))
                    {
                        throw LocalException("AronAbstractTypeXMLReaderData: The node <" + node.name() + "> has the wrong number of children. Expected: " + std::to_string(size) + " children but it has: " + std::to_string(node.nodes().size()) + ".");
                    }
                }
            }
        }
    }
}
