/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// STD/STL


// Header
#include "AronAbstractTypeXMLReader.h"

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>
#include <ArmarXCore/core/rapidxml/wrapper/RapidXmlReader.h>

#include <RobotAPI/libraries/aron/codegenerator/typeReaders/xmlReader/AronAbstractTypeXMLData.h>
#include <RobotAPI/libraries/aron/types/AronAbstractTypeNavigatorFactory.h>

namespace armarx
{
    namespace aron
    {
        namespace codegeneration
        {
            namespace typereader
            {
                const AronObjectTypeXMLReaderFactory AronAbstractTypeXMLReader::Factory;

                void AronAbstractTypeXMLReader::parse(const std::string& xml)
                {
                    RapidXmlReaderPtr reader = RapidXmlReader::FromXmlString(xml);
                    parse(reader);
                }

                void AronAbstractTypeXMLReader::parseFile(const std::string& filename)
                {
                    RapidXmlReaderPtr reader = RapidXmlReader::FromFile(filename);
                    parse(reader);
                }

                void AronAbstractTypeXMLReader::parseFile(const std::filesystem::path& file)
                {
                    RapidXmlReaderPtr reader = RapidXmlReader::FromFile(file.string());
                    parse(reader);
                }

                void AronAbstractTypeXMLReader::parse(const RapidXmlReaderPtr& reader)
                {
                    RapidXmlReaderNode root = reader->getRoot();

                    // Check Header
                    AronAbstractTypeXMLReaderData::EnforceTagName(root, AronAbstractTypeXMLReaderData::ARON_TYPE_DEFINITION_TAG);

                    AronAbstractTypeXMLReaderData::CheckMaxChildSize(root, 4);
                    std::vector<RapidXmlReaderNode> children = root.nodes();

                    int includes_index = -1;
                    int serializers_index = -1;
                    int use_types_index = -1;
                    int produce_types_index = -1;

                    int index = 0;
                    for (const auto& child : children)
                    {
                        if (AronAbstractTypeXMLReaderData::HasTagName(child, AronAbstractTypeXMLReaderData::ARON_INCLUDES_TAG))
                        {
                            if (includes_index != -1)
                            {
                                throw LocalException("AronAbstractTypeXMLReader: Already found an include tag. Please check the xml file.");
                            }
                            else
                            {
                                includes_index = index;
                            }
                        }
                        else if (AronAbstractTypeXMLReaderData::HasTagName(child, AronAbstractTypeXMLReaderData::ARON_SERIALIZERS_TAG))
                        {
                            if (serializers_index != -1)
                            {
                                throw LocalException("AronAbstractTypeXMLReader: Already found a serializers tag. Please check the xml file.");
                            }
                            else
                            {
                                serializers_index = index;
                            }
                        }
                        else if (AronAbstractTypeXMLReaderData::HasTagName(child, AronAbstractTypeXMLReaderData::ARON_USE_TYPES_TAG))
                        {
                            if (use_types_index != -1)
                            {
                                throw LocalException("AronAbstractTypeXMLReader: Already found an use type tag. Please check the xml file.");
                            }
                            else
                            {
                                use_types_index = index;
                            }
                        }
                        else if (AronAbstractTypeXMLReaderData::HasTagName(child, AronAbstractTypeXMLReaderData::ARON_PRODUCE_TYPES_TAG))
                        {
                            if (produce_types_index != -1)
                            {
                                throw LocalException("AronAbstractTypeXMLReader: Already found an generate type tag. Please check the xml file.");
                            }
                            else
                            {
                                produce_types_index = index;
                            }
                        }
                        index++;
                    }

                    // 1. Check includes
                    if (includes_index != -1)
                    {
                        std::vector<RapidXmlReaderNode> includes = children[includes_index].nodes();
                        for (const auto& include : includes)
                        {
                            this->additionalIncludes.push_back(readInclude(include));
                        }
                    }

                    // 2. Check Serializers
                    if (serializers_index != -1)
                    {
                        AronAbstractTypeXMLReaderData::CheckMaxChildSize(children[serializers_index], 2);
                        std::vector<RapidXmlReaderNode> serializers = children[serializers_index].nodes();

                        int writers_index = -1;
                        int readers_index = -1;

                        index = 0;
                        for (const auto& serializer : serializers)
                        {
                            if (AronAbstractTypeXMLReaderData::HasTagName(serializer, AronAbstractTypeXMLReaderData::ARON_WRITERS_TAG))
                            {
                                if (writers_index != -1)
                                {
                                    throw LocalException("AronAbstractTypeXMLReader: Already found an writers tag. Please check the xml file.");
                                }
                                else
                                {
                                    writers_index = index;
                                }
                            }
                            else if (AronAbstractTypeXMLReaderData::HasTagName(serializer, AronAbstractTypeXMLReaderData::ARON_READERS_TAG))
                            {
                                if (readers_index != -1)
                                {
                                    throw LocalException("AronAbstractTypeXMLReader: Already found a readers tag. Please check the xml file.");
                                }
                                else
                                {
                                    readers_index = index;
                                }
                            }
                            index++;
                        }

                        if (writers_index != -1)
                        {
                            for (const auto& serializer : serializers[writers_index].nodes())
                            {
                                this->writers.insert(readWriter(serializer));
                            }
                        }
                        if (readers_index != -1)
                        {
                            for (const auto& serializer : serializers[readers_index].nodes())
                            {
                                this->readers.insert(readReader(serializer));
                            }
                        }
                    }

                    // 3. Check UseTypes
                    if (use_types_index != -1)
                    {
                        for (const auto& receiveType : children[use_types_index].nodes())
                        {
                            this->useTypes.push_back(readUseType(receiveType));
                        }

                        // Add UseTypes to Factory
                        std::vector<std::string> useObjectNames;
                        for (const auto& useType : useTypes)
                        {
                            useObjectNames.push_back(useType.typeName);
                        }
                        Factory.AddTopLevelObjectNames(useObjectNames);
                    }

                    // 4. Check ProduceTypes
                    if (produce_types_index != -1)
                    {
                        for (const auto& produceType : children[produce_types_index].nodes())
                        {
                            readGenerateType(produceType);
                        }
                        generateTypes = Factory.GetAllProducedTopLevelObjects();
                    }
                }

                std::string AronAbstractTypeXMLReader::readInclude(const RapidXmlReaderNode& node) const
                {
                    AronAbstractTypeXMLReaderData::EnforceTagName(node, AronAbstractTypeXMLReaderData::ARON_INCLUDE_TAG);
                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_INCLUDE_ATTRIBUTE_NAME);
                    const std::string include = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_INCLUDE_ATTRIBUTE_NAME.c_str());
                    return include;
                }

                AronWriterInfo AronAbstractTypeXMLReader::readWriter(const RapidXmlReaderNode& node) const
                {
                    AronAbstractTypeXMLReaderData::EnforceTagName(node, AronAbstractTypeXMLReaderData::ARON_WRITER_TAG);

                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_METHOD_ATTRIBUTE_NAME);
                    const std::string methodname = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_METHOD_ATTRIBUTE_NAME.c_str());

                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_RETURN_ATTRIBUTE_NAME);
                    const std::string returntype = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_RETURN_ATTRIBUTE_NAME.c_str());

                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_WRITER_ATTRIBUTE_NAME);
                    const std::string writer = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_WRITER_ATTRIBUTE_NAME.c_str());

                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_INCLUDE_ATTRIBUTE_NAME);
                    const std::string include = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_INCLUDE_ATTRIBUTE_NAME.c_str());

                    AronWriterInfo info;
                    info.methodName = methodname;
                    info.returnType = returntype;
                    info.writerClassType = writer;
                    info.include = include;

                    return info;
                }

                AronReaderInfo AronAbstractTypeXMLReader::readReader(const RapidXmlReaderNode& node) const
                {
                    AronAbstractTypeXMLReaderData::EnforceTagName(node, AronAbstractTypeXMLReaderData::ARON_READER_TAG);

                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_METHOD_ATTRIBUTE_NAME);
                    const std::string methodname = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_METHOD_ATTRIBUTE_NAME.c_str());

                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_ARGUMENT_TYPE_ATTRIBUTE_NAME);
                    const std::string returntype = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_ARGUMENT_TYPE_ATTRIBUTE_NAME.c_str());

                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_READER_ATTRIBUTE_NAME);
                    const std::string reader = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_READER_ATTRIBUTE_NAME.c_str());

                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_INCLUDE_ATTRIBUTE_NAME);
                    const std::string include = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_INCLUDE_ATTRIBUTE_NAME.c_str());

                    AronReaderInfo info;
                    info.methodName = methodname;
                    info.argumentType = returntype;
                    info.readerClassType = reader;
                    info.include = include;

                    return info;
                }

                AronUseTypeInfo AronAbstractTypeXMLReader::readUseType(const RapidXmlReaderNode& node) const
                {
                    AronAbstractTypeXMLReaderData::EnforceTagName(node, AronAbstractTypeXMLReaderData::ARON_USE_TYPE_TAG);

                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_INCLUDE_ATTRIBUTE_NAME);
                    const std::string include = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_INCLUDE_ATTRIBUTE_NAME.c_str());

                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_NAME_ATTRIBUTE_NAME);
                    const std::string name = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_NAME_ATTRIBUTE_NAME.c_str());

                    AronUseTypeInfo info;
                    info.typeName = name;
                    info.include = include;

                    return info;
                }

                types::AronObjectTypeNavigatorPtr AronAbstractTypeXMLReader::readGenerateType(const RapidXmlReaderNode& node) const
                {
                    AronAbstractTypeXMLReaderData::EnforceTagName(node, AronAbstractTypeXMLReaderData::ARON_PRODUCE_TYPE_TAG);
                    return types::AronObjectTypeNavigator::DynamicCast(Factory.createSpecific(node, AronNavigatorPath()));
                }
            }
        }
    }
}
