/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// STD/STL

// Boost
#include <boost/algorithm/string.hpp>

// Header
#include "AronAbstractTypeXMLReaderFactory.h"

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>

#include <RobotAPI/libraries/aron/codegenerator/typeReaders/xmlReader/AronAbstractTypeXMLData.h>

namespace armarx
{
    namespace aron
    {
        namespace codegeneration
        {
            namespace typereader
            {
                std::string AronAbstractTypeXMLReaderFactory::CurrentTopLevelObjectName = "";
                std::vector<std::pair<std::string, std::vector<std::string>>> AronAbstractTypeXMLReaderFactory::AllTopLevelObjectNames = {};
                std::vector<std::string> AronAbstractTypeXMLReaderFactory::AllLocalObjectNames = {};
                std::vector<std::pair<types::AronObjectTypeNavigatorPtr, std::vector<types::AronObjectTypeNavigatorPtr>>> AronAbstractTypeXMLReaderFactory::AllTopLevelProducedObjects = {};
                std::vector<types::AronObjectTypeNavigatorPtr> AronAbstractTypeXMLReaderFactory::AllLocalProducedObjects = {};

                const std::map<std::string, AronAbstractTypeXMLReaderFactoryPtr> AronAbstractTypeXMLReaderFactory::Factories =
                {
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    {AronAbstractTypeXMLReaderData::ARON_PRODUCE_##capsType##_TAG, AronAbstractTypeXMLReaderFactoryPtr(new Aron##upperType##TypeXMLReaderFactory())}, \

                    HANDLE_CONTAINER_TYPES
                    HANDLE_EXTERNAL_TYPES
                    HANDLE_COMPLEX_TYPES
#undef RUN_ARON_MACRO


#define RUN_ARON_MACRO(cppType, upperType, lowerType, capsType) \
    {AronAbstractTypeXMLReaderData::ARON_PRODUCE_##capsType##_TAG, AronAbstractTypeXMLReaderFactoryPtr(new AronPrimitiveTypeXMLReaderFactory<Aron##upperType##Type, Aron##upperType, cppType>())}, \

                    HANDLE_PRIMITIVE_TYPES_WITH_CPP_TYPE
#undef RUN_ARON_MACRO
                };

                types::AronAbstractTypeNavigatorPtr AronAbstractTypeXMLReaderFactory::create(const RapidXmlReaderNode& node, const AronNavigatorPath& path) const
                {
                    const std::string tag = boost::algorithm::to_lower_copy(node.name());
                    if (tag == "")
                    {
                        throw LocalException("AronAbstractTypeXMLReaderFactory: An node is empty. Cannot find facory for empty nodes.");
                    }

                    auto factory_iterator = Factories.find(tag);
                    if (factory_iterator == Factories.end())
                    {
                        throw LocalException("AronAbstractTypeXMLReaderFactory: Cannot find the desired factory for tag '" + tag + "' (Note that we convert everything to lowercase). Cannot create navigator!");
                    }
                    return factory_iterator->second->createSpecific(node, path);
                }

                std::vector<std::pair<types::AronObjectTypeNavigatorPtr, std::vector<types::AronObjectTypeNavigatorPtr>>> AronAbstractTypeXMLReaderFactory::GetAllProducedTopLevelObjects()
                {
                    return AllTopLevelProducedObjects;
                }

                void AronAbstractTypeXMLReaderFactory::AddTopLevelObjectNames(const std::vector<std::string>& names)
                {
                    for (const auto& name : names)
                    {
                        AddTopLevelObjectName(name);
                    }
                }

                void AronAbstractTypeXMLReaderFactory::AddTopLevelObjectName(const std::string& name)
                {
                    AllTopLevelObjectNames.push_back(std::make_pair(name, std::vector<std::string>()));
                }

                // External
                types::AronAbstractTypeNavigatorPtr AronExternalTypeXMLReaderFactory::createSpecific(const RapidXmlReaderNode& node, const AronNavigatorPath& path) const
                {
                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_NAME_ATTRIBUTE_NAME);
                    const std::string name = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_NAME_ATTRIBUTE_NAME.c_str());

                    // check whether type exists
                    for (const auto& [x, _] : AllTopLevelObjectNames)
                    {
                        if (name == x)
                        {
                            types::AronExternalTypeNavigatorPtr aronExternalType = types::AronExternalTypeNavigatorPtr(new types::AronExternalTypeNavigator(name, path));
                            return aronExternalType;
                        }
                    }

                    // check whether type locally exists
                    if (std::find(AllLocalObjectNames.begin(), AllLocalObjectNames.end(), name) != AllLocalObjectNames.end())
                    {
                        types::AronExternalTypeNavigatorPtr aronExternalType = types::AronExternalTypeNavigatorPtr(new types::AronExternalTypeNavigator(name, path));
                        return aronExternalType;
                    }

                    throw LocalException("AronExternalTypeXMLReaderFactory: A name could not be found. Perhaps you need to change the order of your types or you have to add another UseType!");
                }

                // Object
                types::AronAbstractTypeNavigatorPtr AronObjectTypeXMLReaderFactory::createSpecific(const RapidXmlReaderNode& node, const AronNavigatorPath& path) const
                {
                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_NAME_ATTRIBUTE_NAME);
                    const std::string name = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_NAME_ATTRIBUTE_NAME.c_str());

                    if (AronAbstractTypeXMLReaderData::HasTagName(node, AronAbstractTypeXMLReaderData::ARON_PRODUCE_TYPE_TAG))
                    {
                        // we are in a top level object
                        // set the new name
                        CurrentTopLevelObjectName = name;
                        AllLocalObjectNames.clear();
                        AllLocalProducedObjects.clear();
                    }
                    else if (name.find("::") != std::string::npos)
                    {
                        throw LocalException("AronObjectTypeXMLReaderFactory: A local class must not have a namespace!");
                    }

                    types::AronObjectTypeNavigatorPtr aronObjectType = types::AronObjectTypeNavigatorPtr(new types::AronObjectTypeNavigator(name, path));

                    for (const RapidXmlReaderNode& objectChild : node.nodes())
                    {
                        AronAbstractTypeXMLReaderData::EnforceTagName(objectChild, AronAbstractTypeXMLReaderData::ARON_OBJECT_CHILD_TAG);
                        AronAbstractTypeXMLReaderData::EnforceChildSize(objectChild, 1);

                        AronAbstractTypeXMLReaderData::EnforceAttribute(objectChild, AronAbstractTypeXMLReaderData::ARON_KEY_ATTRIBUTE_NAME);
                        const std::string key = objectChild.attribute_value(AronAbstractTypeXMLReaderData::ARON_KEY_ATTRIBUTE_NAME.c_str());

                        std::vector<RapidXmlReaderNode> children = objectChild.nodes();
                        types::AronAbstractTypeNavigatorPtr childNavigator = create(children[0], AronNavigatorPath(path, key));
                        aronObjectType->addElement(key, childNavigator);
                    }

                    if (AronAbstractTypeXMLReaderData::HasTagName(node, AronAbstractTypeXMLReaderData::ARON_PRODUCE_TYPE_TAG))
                    {
                        // Finishing top level object
                        AllTopLevelObjectNames.push_back(std::make_pair(name, AllLocalObjectNames));
                        AllTopLevelProducedObjects.push_back(std::make_pair(aronObjectType, AllLocalProducedObjects));
                    }
                    else
                    {
                        std::cout << "add low level " << name;
                        AllLocalObjectNames.push_back(aronObjectType->getFullCppTypename());
                        AllLocalProducedObjects.push_back(aronObjectType);
                    }

                    return aronObjectType;
                }

                // List type (List)
                types::AronAbstractTypeNavigatorPtr AronListTypeXMLReaderFactory::createSpecific(const RapidXmlReaderNode& node, const AronNavigatorPath& path) const
                {
                    if (node.has_attribute(AronAbstractTypeXMLReaderData::ARON_TYPE_ATTRIBUTE_NAME.c_str()))
                    {
                        AronAbstractTypeXMLReaderData::EnforceChildSize(node, 0);
                        const std::string acceptedType = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_TYPE_ATTRIBUTE_NAME.c_str());

                        return types::AronAbstractTypeNavigator::GetAronAbstractTypePtrByName(acceptedType);
                    }
                    else
                    {
                        AronAbstractTypeXMLReaderData::EnforceChildSize(node, 1);
                        std::vector<RapidXmlReaderNode> nodeChildren = node.nodes();
                        const RapidXmlReaderNode listTypeDeclarationNode = nodeChildren[0];

                        AronAbstractTypeXMLReaderData::EnforceTagName(listTypeDeclarationNode, AronAbstractTypeXMLReaderData::ARON_TYPE_TAG);

                        AronAbstractTypeXMLReaderData::EnforceChildSize(listTypeDeclarationNode, 1);
                        std::vector<RapidXmlReaderNode> listTypeNodeChildren = listTypeDeclarationNode.nodes();
                        const RapidXmlReaderNode typeNode = listTypeNodeChildren[0];
                        types::AronAbstractTypeNavigatorPtr type = create(typeNode, AronNavigatorPath(path, "type"));

                        return types::AronAbstractTypeNavigatorPtr(new types::AronListTypeNavigator(type, path));
                    }
                }

                // List type (Dict)
                types::AronAbstractTypeNavigatorPtr AronDictTypeXMLReaderFactory::createSpecific(const RapidXmlReaderNode& node, const AronNavigatorPath& path) const
                {
                    AronAbstractTypeXMLReaderData::EnforceChildSize(node, 1);
                    std::vector<RapidXmlReaderNode> nodeChildren = node.nodes();
                    const RapidXmlReaderNode dictTypeDeclarationNode = nodeChildren[0];

                    if (dictTypeDeclarationNode.has_attribute(AronAbstractTypeXMLReaderData::ARON_TYPE_ATTRIBUTE_NAME.c_str()))
                    {
                        AronAbstractTypeXMLReaderData::EnforceChildSize(dictTypeDeclarationNode, 0);
                        const std::string acceptedType = dictTypeDeclarationNode.attribute_value(AronAbstractTypeXMLReaderData::ARON_TYPE_ATTRIBUTE_NAME.c_str());
                        return types::AronAbstractTypeNavigator::GetAronAbstractTypePtrByName(acceptedType);
                    }
                    else
                    {
                        AronAbstractTypeXMLReaderData::EnforceTagName(dictTypeDeclarationNode, AronAbstractTypeXMLReaderData::ARON_TYPE_TAG);

                        AronAbstractTypeXMLReaderData::EnforceChildSize(dictTypeDeclarationNode, 1);
                        std::vector<RapidXmlReaderNode> dictTypeNodeChildren = dictTypeDeclarationNode.nodes();
                        const RapidXmlReaderNode typeNode = dictTypeNodeChildren[0];
                        types::AronAbstractTypeNavigatorPtr type = create(typeNode, AronNavigatorPath(path, "type"));
                        return types::AronAbstractTypeNavigatorPtr(new types::AronDictTypeNavigator(type, path));
                    }
                }

                // List types (Tuple)
                types::AronAbstractTypeNavigatorPtr AronTupleTypeXMLReaderFactory::createSpecific(const RapidXmlReaderNode& node, const AronNavigatorPath& path) const
                {
                    std::vector<types::AronAbstractTypeNavigatorPtr> types;
                    std::vector<RapidXmlReaderNode> nodeChildren = node.nodes();

                    if (nodeChildren.size() < 2)
                    {
                        throw LocalException("AronTupleTypeXMLReaderFactory: A tuple must have at least two types.");
                    }

                    unsigned int i = 0;
                    for (const RapidXmlReaderNode& tupleTypeDeclarationNode : nodeChildren)
                    {
                        AronAbstractTypeXMLReaderData::EnforceTagName(tupleTypeDeclarationNode, AronAbstractTypeXMLReaderData::ARON_TYPE_TAG);

                        if (tupleTypeDeclarationNode.has_attribute(AronAbstractTypeXMLReaderData::ARON_TYPE_ATTRIBUTE_NAME.c_str()))
                        {
                            AronAbstractTypeXMLReaderData::EnforceChildSize(tupleTypeDeclarationNode, 0);

                            const std::string typeName = tupleTypeDeclarationNode.attribute_value(AronAbstractTypeXMLReaderData::ARON_TYPE_ATTRIBUTE_NAME.c_str());
                            types::AronAbstractTypeNavigatorPtr type = types::AronAbstractTypeNavigator::GetAronAbstractTypePtrByName(typeName);
                        }
                        else
                        {
                            AronAbstractTypeXMLReaderData::EnforceChildSize(tupleTypeDeclarationNode, 1);

                            std::vector<RapidXmlReaderNode> typeNodeChildren = tupleTypeDeclarationNode.nodes();
                            const RapidXmlReaderNode typeNode = typeNodeChildren[0];

                            types::AronAbstractTypeNavigatorPtr type = create(typeNode, AronNavigatorPath(path, "<" + std::to_string(i++) + ">"));
                            types.push_back(type);
                        }
                    }
                    return types::AronAbstractTypeNavigatorPtr(new types::AronTupleTypeNavigator(types, path));
                }

                // Complex type (IVTCByteImage)
                types::AronAbstractTypeNavigatorPtr AronIVTCByteImageTypeXMLReaderFactory::createSpecific(const RapidXmlReaderNode& node, const AronNavigatorPath& path) const
                {
                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_WIDTH_ATTRIBUTE_NAME);
                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_HEIGHT_ATTRIBUTE_NAME);
                    AronAbstractTypeXMLReaderData::EnforceAttribute(node, AronAbstractTypeXMLReaderData::ARON_TYPE_ATTRIBUTE_NAME);

                    int width = std::stoi(node.attribute_value(AronAbstractTypeXMLReaderData::ARON_WIDTH_ATTRIBUTE_NAME.c_str()));
                    int height = std::stoi(node.attribute_value(AronAbstractTypeXMLReaderData::ARON_HEIGHT_ATTRIBUTE_NAME.c_str()));
                    std::string type = node.attribute_value(AronAbstractTypeXMLReaderData::ARON_TYPE_ATTRIBUTE_NAME.c_str());;


                    return types::AronAbstractTypeNavigatorPtr(new types::AronIVTCByteImageTypeNavigator(width, height, type, path));
                }

                // Complex type (EigenMatrix)
                types::AronAbstractTypeNavigatorPtr AronEigenMatrixTypeXMLReaderFactory::createSpecific(const RapidXmlReaderNode& node, const AronNavigatorPath& path) const
                {
                    throw LocalException("AronTupleTypeXMLReaderFactory: Not implemented yet.");
                }

                // Complex type (OpenCVMat)
                types::AronAbstractTypeNavigatorPtr AronOpenCVMatTypeXMLReaderFactory::createSpecific(const RapidXmlReaderNode& node, const AronNavigatorPath& path) const
                {
                    throw LocalException("AronTupleTypeXMLReaderFactory: Not implemented yet.");
                }

                // Complex type (PCLPointcloud)
                types::AronAbstractTypeNavigatorPtr AronPCLPointcloudTypeXMLReaderFactory::createSpecific(const RapidXmlReaderNode& node, const AronNavigatorPath& path) const
                {
                    throw LocalException("AronTupleTypeXMLReaderFactory: Not implemented yet.");
                }
            }
        }
    }
}
