/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>

// Header
#include "AronDataNavigatorReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/data/AronListDataNavigator.h>
#include <RobotAPI/libraries/aron/data/AronPrimitiveDataNavigator.h>

using namespace std;

namespace armarx
{
    namespace aron
    {
        namespace io
        {

            AronDataNavigatorReader::AronDataNavigatorReader(const AronDataPtr& aron) :
                AronDataReader<AronDataPtr>(aron),
                input(data::AronDataNavigator::FromAronPtr(aron, AronNavigatorPath())),
                readInitialStart(false)
            {

            }

            bool AronDataNavigatorReader::readStartList()
            {
                if (!readInitialStart)
                {
                    data::AronListDataNavigatorPtr l = data::AronListDataNavigator::DynamicCast(input);
                    AronDataNavigatorReaderTokenPtr token = AronDataNavigatorReaderTokenPtr(new AronDataNavigatorReaderToken(l));
                    stack.push(token);

                    readInitialStart = true;
                    return true;
                }
                AronDataNavigatorReaderTokenPtr lastToken = stack.top();
                data::AronListDataNavigatorPtr list = data::AronDataNavigator::DynamicCast<data::AronListDataNavigator>(lastToken->getElementAndIncreaseCnt());

                AronDataNavigatorReaderTokenPtr newToken = AronDataNavigatorReaderTokenPtr(new AronDataNavigatorReaderToken(list));
                stack.push(newToken);
                return true;
            }

            bool AronDataNavigatorReader::readStartDict()
            {
                if (!readInitialStart)
                {
                    data::AronDictDataNavigatorPtr obj = data::AronDictDataNavigator::DynamicCast(input);
                    AronDataNavigatorReaderTokenPtr token = AronDataNavigatorReaderTokenPtr(new AronDataNavigatorReaderToken(obj));
                    stack.push(token);

                    readInitialStart = true;
                    return true;
                }
                AronDataNavigatorReaderTokenPtr lastToken = stack.top();
                data::AronDictDataNavigatorPtr dict = data::AronDataNavigator::DynamicCast<data::AronDictDataNavigator>(lastToken->getElementAndIncreaseCnt());

                AronDataNavigatorReaderTokenPtr newToken = AronDataNavigatorReaderTokenPtr(new AronDataNavigatorReaderToken(dict));
                stack.push(newToken);
                return true;
            }

            bool AronDataNavigatorReader::readEndList()
            {
                AronDataNavigatorReaderTokenPtr token = stack.top();
                if (token->getDescriptor() != AronDataDescriptor::eAronListData)
                {
                    throw LocalException("AronDataNavigatorReader: A token in the stack has the wrong type. Expected List!");
                }

                if (token->checkedAllChildren())
                {
                    stack.pop();
                    return true;
                }
                return false;
            }

            bool AronDataNavigatorReader::readEndDict()
            {
                AronDataNavigatorReaderTokenPtr token = stack.top();
                if (token->getDescriptor() != AronDataDescriptor::eAronDictData)
                {
                    throw LocalException("AronDataNavigatorReader: A token in the stack has the wrong type. Expected Dict!");
                    return false;
                }

                if (token->checkedAllChildren())
                {
                    stack.pop();
                    return true;
                }
                return false;
            }

            // Complex Data
            unsigned char* AronDataNavigatorReader::readBlob()
            {
                AronDataNavigatorReaderTokenPtr token = stack.top();
                data::AronDataNavigatorPtr nav = token->getElementAndIncreaseCnt();
                data::AronBlobDataNavigatorPtr casted = data::AronBlobDataNavigator::DynamicCast(nav);

                return casted->getData();
            }

            // Read primitives
#define RUN_ARON_MACRO(cppType, upperType, lowerType, capsType) \
    cppType AronDataNavigatorReader::read##upperType() \
    { \
        AronDataNavigatorReaderTokenPtr token = stack.top(); \
        data::AronDataNavigatorPtr nav = token->getElementAndIncreaseCnt(); \
        data::AronPrimitiveDataNavigatorPtr<Aron##upperType, cppType> casted = data::AronPrimitiveDataNavigator<Aron##upperType, cppType>::DynamicCast(nav); \
        return casted->getValue(); \
    }

            HANDLE_PRIMITIVE_TYPES_WITH_CPP_TYPE
#undef RUN_ARON_MACRO

            // ReadKey
            std::string AronDataNavigatorReader::readKey()
            {
                AronDataNavigatorReaderTokenPtr token = stack.top();
                return token->getCurrentKey();
            }

            bool AronDataNavigatorReader::readMember(const std::string& k)
            {
                AronDataNavigatorReaderTokenPtr token = stack.top();
                token->addCurrentKey(k);
                return true;
            }
        }
    }
}
