/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>

// Header
#include "AronDataNavigatorWriter.h"

// ArmarX
#include <RobotAPI/libraries/aron/data/AronListDataNavigator.h>
#include <RobotAPI/libraries/aron/data/AronPrimitiveDataNavigator.h>

#include <RobotAPI/libraries/aron/types/AronListTypeNavigator.h>
#include <RobotAPI/libraries/aron/types/AronObjectTypeNavigator.h>
#include <RobotAPI/libraries/aron/types/AronTupleTypeNavigator.h>
#include <RobotAPI/libraries/aron/types/AronPrimitiveTypeNavigator.h>

using namespace std;

namespace armarx
{
    namespace aron
    {
        namespace io
        {

            AronDataNavigatorWriter::AronDataNavigatorWriter() :
                wroteInitialStartContainer(false)
            {

            }

            bool AronDataNavigatorWriter::writeStartList()
            {
                AronNavigatorPath path;
                if (!wroteInitialStartContainer)
                {
                    wroteInitialStartContainer = true;
                }
                else
                {
                    AronDataNavigatorWriterTokenPtr token = stack.top();
                    path = AronNavigatorPath(token->getData()->getPath(), token->getAccessor());
                }

                data::AronDataNavigatorPtr data = data::AronDataNavigatorPtr(new data::AronListDataNavigator(path));
                AronDataNavigatorWriterTokenPtr token2 = AronDataNavigatorWriterTokenPtr(new AronDataNavigatorWriterToken(data));
                stack.push(token2);
                return true;
            }

            bool AronDataNavigatorWriter::writeStartDict()
            {
                AronNavigatorPath path;
                if (!wroteInitialStartContainer)
                {
                    wroteInitialStartContainer = true;
                }
                else
                {
                    AronDataNavigatorWriterTokenPtr token = stack.top();
                    path = AronNavigatorPath(token->getData()->getPath(), token->getAccessor());
                }
                data::AronDataNavigatorPtr data = data::AronDataNavigatorPtr(new data::AronDictDataNavigator(path));
                AronDataNavigatorWriterTokenPtr token = AronDataNavigatorWriterTokenPtr(new AronDataNavigatorWriterToken(data));
                stack.push(token);
                return true;
            }

            bool AronDataNavigatorWriter::writeEndList()
            {
                AronDataNavigatorWriterTokenPtr token = stack.top();
                lastRemovedElement = token->getData();
                stack.pop();

                if (stack.size() > 0)
                {
                    AronDataNavigatorWriterTokenPtr prevToken = stack.top();
                    prevToken->addElement(lastRemovedElement);
                }
                return true;
            }

            bool AronDataNavigatorWriter::writeEndDict()
            {
                AronDataNavigatorWriterTokenPtr token = stack.top();
                lastRemovedElement = token->getData();
                stack.pop();

                if (stack.size() > 0)
                {
                    AronDataNavigatorWriterTokenPtr prevToken = stack.top();
                    prevToken->addElement(lastRemovedElement);
                }
                return true;
            }

            // Complex Types
            bool AronDataNavigatorWriter::writeBlob(const unsigned int elements, const unsigned char* data)
            {
                AronDataNavigatorWriterTokenPtr token = stack.top();
                data::AronBlobDataNavigatorPtr aron(new data::AronBlobDataNavigator(AronNavigatorPath()));
                aron->setData(elements, data);
                token->addElement(aron);

                return true;
            }

#define RUN_ARON_MACRO(cppType, upperType, lowerType, capsType) \
    bool AronDataNavigatorWriter::write##upperType(const cppType& x) \
    { \
        AronDataNavigatorWriterTokenPtr token = stack.top(); \
        data::AronPrimitiveDataNavigatorPtr<Aron##upperType, cppType> aron(new data::AronPrimitiveDataNavigator<Aron##upperType, cppType>(AronNavigatorPath())); \
        aron->setValue(x); \
        token->addElement(aron); \
        return true; \
    }

            HANDLE_PRIMITIVE_TYPES_WITH_CPP_TYPE
#undef RUN_ARON_MACRO

            bool AronDataNavigatorWriter::writeKey(const std::string& k)
            {
                AronDataNavigatorWriterTokenPtr token = stack.top();
                token->setCurrentKey(k);
                return true;
            }

            AronDataPtr AronDataNavigatorWriter::getResult() const
            {
                if (lastRemovedElement == nullptr)
                {
                    throw LocalException("AronDataNavigatorWriter: Could not return NULL. Perhaps you never called a writeEnd* method?");
                }
                return lastRemovedElement->getResult();
            }
        }
    }
}
