/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// Header
#include "AronDataNavigatorWriterToken.h"

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>

namespace armarx
{
    namespace aron
    {
        namespace io
        {
            // constructor
            AronDataNavigatorWriterToken::AronDataNavigatorWriterToken(const data::AronDataNavigatorPtr& data) :
                data(data),
                currentIndex(0),
                currentKey("")
            {

            }

            AronDataDescriptor AronDataNavigatorWriterToken::getDescriptor() const
            {
                return data->getDescriptor();
            }

            void AronDataNavigatorWriterToken::addElement(const data::AronDataNavigatorPtr& n)
            {
                switch (getDescriptor())
                {
                    case eAronDictData:
                    {
                        data::AronDictDataNavigatorPtr casted = data::AronDictDataNavigator::DynamicCast(data);
                        casted->addElement(getCurrentKey(), n);
                        break;
                    }
                    case eAronListData:
                    {
                        data::AronListDataNavigatorPtr casted = data::AronListDataNavigator::DynamicCast(data);
                        casted->addElement(n);
                        currentIndex++;
                        break;
                    }
                    default:
                        throw LocalException("AronDataNavigatorWriterToken: Could not resove a type of a navigator. Allowed are only container and complex types due to performance. The path was: " + data->pathToString());
                }
            }

            data::AronDataNavigatorPtr AronDataNavigatorWriterToken::getData() const
            {
                return data;
            }

            std::string AronDataNavigatorWriterToken::getAccessor() const
            {
                switch (getDescriptor())
                {
                    case eAronDictData:
                        return currentKey;
                    case eAronListData:
                        return std::to_string(currentIndex);
                    default:
                        throw LocalException("AronDataNavigatorWriterToken: Could not resove a type of a navigator. Allowed are only container and complex types due to performance. The path was: " + data->pathToString());
                }
            }

            // List and Tuple member functions
            unsigned int AronDataNavigatorWriterToken::getCurrentIndex() const
            {
                if (getDescriptor() != AronDataDescriptor::eAronListData)
                {
                    throw LocalException("AronDataNavigatorWriterToken: Cant get index of a non-list token. The path was: " + data->pathToString());
                }
                return currentIndex;
            }

            // Dict and object member functions
            void AronDataNavigatorWriterToken::setCurrentKey(const std::string& k)
            {
                if (getDescriptor() != AronDataDescriptor::eAronDictData)
                {
                    throw LocalException("AronDataNavigatorWriterToken: Cant set key of a non-dict token. The path was: " + data->pathToString());
                }
                currentKey = k;
            }

            std::string AronDataNavigatorWriterToken::getCurrentKey() const
            {
                if (getDescriptor() != AronDataDescriptor::eAronDictData)
                {
                    throw LocalException("AronDataNavigatorWriterToken: Cant get key of a non-dict token. The path was: " + data->pathToString());
                }
                return currentKey;
            }
        }
    }
}
