/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::aron
 * @author     Simon Ottenhaus ( simon dot ottenhaus at kit dot edu )
 * @date       2019
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE RobotAPI::ArmarXLibraries::aron

#define ARMARX_BOOST_TEST

// STD/STL
#include <iostream>
#include <cstdlib>
#include <ctime>

// IVT
#include <Image/ByteImage.h>

// Eigen
#include <Eigen/Core>

// Boost
#include <boost/algorithm/string.hpp>

// Test
#include <RobotAPI/Test.h>

// ArmarX
#include <ArmarXCore/libraries/cppgen/CppMethod.h>
#include <ArmarXCore/libraries/cppgen/CppClass.h>

// Generated File
#include "aron/NaturalIK.aron.generated.h"
#include "aron/ListTest.aron.generated.h"
#include "aron/DictTest.aron.generated.h"
#include "aron/PrimitiveTest.aron.generated.h"
#include "aron/ObjectTest.aron.generated.h"
#include "aron/ImageTest.aron.generated.h"


using namespace armarx;
using namespace aron;

void initialize_random()
{
    std::srand(std::time(nullptr));
}

int generateRandom(int max, int min)
{
    int random = (std::rand() % max) + min;
    return random;
}

std::string generateRandomWord()
{
    vector<string> words = {"Lorem", "ipsum", "dolor", "sit", "amet", "consetetur", "sadipscing", "elitr"
                            "sed", "diam", "nonumy", "eirmod", "tempor", "invidunt", "ut", "labore", "et"
                            "dolore", "magna", "aliquyam", "eratsed"
                           };

    int i = generateRandom(words.size(), 0);
    return words[i];
}

std::vector<unsigned char> generateRandomBlob(unsigned int size)
{
    std::vector<unsigned char> new_blob(size, 0);
    for (unsigned int i = 0; i < size; ++i)
    {
        new_blob[i] = (generateRandom(127, 0));
    }
    return new_blob;
}

AronDataPtr generateAronDataFromType(const AronAbstractTypePtr& type)
{
    {
        AronObjectTypePtr t = AronObjectTypePtr::dynamicCast(type);
        if (t)
        {
            AronDictPtr d = AronDictPtr(new AronDict());
            for (const auto& [k, tt] : t->elementTypes)
            {
                d->elements[k] = generateAronDataFromType(tt);
            }
            return d;
        }
    }
    {
        AronDictTypePtr t = AronDictTypePtr::dynamicCast(type);
        if (t)
        {
            return new AronDict();
        }
    }
    {
        AronListTypePtr t = AronListTypePtr::dynamicCast(type);
        if (t)
        {
            return new AronList();
        }
    }
    {
        AronTupleTypePtr t = AronTupleTypePtr::dynamicCast(type);
        if (t)
        {
            return new AronList();
        }
    }
    {
        AronEigenMatrixTypePtr t = AronEigenMatrixTypePtr::dynamicCast(type);
        if (t)
        {
            return new AronBlob();
        }
    }
    {
        AronIntTypePtr t = AronIntTypePtr::dynamicCast(type);
        if (t)
        {
            return new AronInt();
        }
    }
    {
        AronLongTypePtr t = AronLongTypePtr::dynamicCast(type);
        if (t)
        {
            return new AronLong();
        }
    }
    {
        AronFloatTypePtr t = AronFloatTypePtr::dynamicCast(type);
        if (t)
        {
            return new AronFloat();
        }
    }
    {
        AronDoubleTypePtr t = AronDoubleTypePtr::dynamicCast(type);
        if (t)
        {
            return new AronDouble();
        }
    }
    {
        AronStringTypePtr t = AronStringTypePtr::dynamicCast(type);
        if (t)
        {
            return new AronString();
        }
    }
    {
        AronBoolTypePtr t = AronBoolTypePtr::dynamicCast(type);
        if (t)
        {
            return new AronBool();
        }
    }
    throw LocalException("generateAronDataFromType: No valid type found!");
}

void initializeRandomly(AronDataPtr& data, const AronAbstractTypePtr& type)
{
    BOOST_CHECK_NE(data.get(), nullptr);
    BOOST_CHECK_NE(type.get(), nullptr);

    // Containers
    {
        AronObjectTypePtr objectType = AronObjectTypePtr::dynamicCast(type);
        if (objectType)
        {
            AronDictPtr dict = AronDictPtr::dynamicCast(data);
            BOOST_CHECK_NE(dict.get(), nullptr);
            for (auto& [key, nextData] : dict->elements)
            {
                initializeRandomly(nextData, objectType->elementTypes[key]);
            }
            return;
        }
    }

    {
        AronListTypePtr listType = AronListTypePtr::dynamicCast(type);
        if (listType)
        {
            AronListPtr list = AronListPtr::dynamicCast(data);
            BOOST_CHECK_NE(list.get(), nullptr);

            list->elements.clear();
            int numElements = generateRandom(26, 0);
            for (int i = 0; i < numElements; ++i)
            {
                AronDataPtr newData = generateAronDataFromType(listType->acceptedType);
                initializeRandomly(newData, listType->acceptedType);
                list->elements.push_back(newData);
            }
            return;
        }
    }

    {
        AronDictTypePtr dictType = AronDictTypePtr::dynamicCast(type);
        if (dictType)
        {
            AronDictPtr dict = AronDictPtr::dynamicCast(data);
            BOOST_CHECK_NE(dict.get(), nullptr);

            dict->elements.clear();
            int numElements = generateRandom(26, 0);
            for (int i = 0; i < numElements; ++i)
            {
                std::string key = generateRandomWord();
                AronDataPtr newData = generateAronDataFromType(dictType->acceptedType);
                initializeRandomly(newData, dictType->acceptedType);
                dict->elements[key] = newData;
            }
            return;
        }
    }

    {
        AronTupleTypePtr tupleType = AronTupleTypePtr::dynamicCast(type);
        if (tupleType)
        {
            AronListPtr list = AronListPtr::dynamicCast(data);
            BOOST_CHECK_NE(list.get(), nullptr);
            BOOST_CHECK_EQUAL(list->elements.size(), tupleType->elementTypes.size());

            int i = 0;
            for (auto& nextData : list->elements)
            {
                initializeRandomly(nextData, tupleType->elementTypes[i++]);
            }
            return;
        }
    }

    // Complex
    {
        AronEigenMatrixTypePtr eigenMatrixType = AronEigenMatrixTypePtr::dynamicCast(type);
        if (eigenMatrixType)
        {
            AronBlobPtr blob = AronBlobPtr::dynamicCast(data);
            BOOST_CHECK_NE(blob.get(), nullptr);
            return;
        }

        AronIVTCByteImageTypePtr ivtCByteImageType = AronIVTCByteImageTypePtr::dynamicCast(type);
        if (ivtCByteImageType)
        {
            AronBlobPtr blob = AronBlobPtr::dynamicCast(data);
            BOOST_CHECK_NE(blob.get(), nullptr);

            blob->data = generateRandomBlob(blob->data.size());
            return;
        }
    }


    // Primitives
    {
        AronIntTypePtr primitiveType = AronIntTypePtr::dynamicCast(type);
        if (primitiveType)
        {
            AronIntPtr primitiveData = AronIntPtr::dynamicCast(data);
            BOOST_CHECK_NE(primitiveData.get(), nullptr);

            primitiveData->value = generateRandom(123456789, -123456789);
            return;
        }
    }
    {
        AronLongTypePtr primitiveType = AronLongTypePtr::dynamicCast(type);
        if (primitiveType)
        {
            AronLongPtr primitiveData = AronLongPtr::dynamicCast(data);
            BOOST_CHECK_NE(primitiveData.get(), nullptr);

            primitiveData->value = generateRandom(123456789, -123456789);
            return;
        }
    }
    {
        AronFloatTypePtr primitiveType = AronFloatTypePtr::dynamicCast(type);
        if (primitiveType)
        {
            AronFloatPtr primitiveData = AronFloatPtr::dynamicCast(data);
            BOOST_CHECK_NE(primitiveData.get(), nullptr);

            primitiveData->value = generateRandom(123456789, -123456789);
            return;
        }
    }
    {
        AronDoubleTypePtr primitiveType = AronDoubleTypePtr::dynamicCast(type);
        if (primitiveType)
        {
            AronDoublePtr primitiveData = AronDoublePtr::dynamicCast(data);
            BOOST_CHECK_NE(primitiveData.get(), nullptr);

            primitiveData->value = generateRandom(123456789, -123456789);
            return;
        }
    }
    {
        AronStringTypePtr primitiveType = AronStringTypePtr::dynamicCast(type);
        if (primitiveType)
        {
            AronStringPtr primitiveData = AronStringPtr::dynamicCast(data);
            BOOST_CHECK_NE(primitiveData.get(), nullptr);

            primitiveData->value = generateRandomWord();
            return;
        }
    }
    {
        AronBoolTypePtr primitiveType = AronBoolTypePtr::dynamicCast(type);
        if (primitiveType)
        {
            AronBoolPtr primitiveData = AronBoolPtr::dynamicCast(data);
            BOOST_CHECK_NE(primitiveData.get(), nullptr);

            primitiveData->value = generateRandom(1, 0);
            return;
        }
    }
    throw LocalException("initializeRandomly: No valid type found!");
}

template <typename T>
void runTestWithInstances(T& k, T& k2)
{
    AronAbstractTypePtr k_type = k.getAronAbstractType();
    BOOST_CHECK_NE(k_type.get(), nullptr);

    AronAbstractTypePtr k2_type = k2.getAronAbstractType();
    BOOST_CHECK_NE(k_type.get(), nullptr);

    AronDataPtr k_aron = k.toAron();
    BOOST_CHECK_NE(k_aron.get(), nullptr);
    initializeRandomly(k_aron, k_type);

    k2.fromAron(k_aron);
    AronDataPtr k2_aron = k2.toAron();
    BOOST_CHECK_NE(k2_aron.get(), nullptr);

    k.fromAron(k2_aron);
    BOOST_CHECK_EQUAL((k == k2), true);
}

BOOST_AUTO_TEST_CASE(AronTests)
{
    initialize_random();

    NaturalIKResult k;
    NaturalIKResult k2;
    runTestWithInstances<NaturalIKResult>(k, k2);

    ListTest l;
    ListTest l2;
    runTestWithInstances<ListTest>(l, l2);

    DictTest d;
    DictTest d2;
    runTestWithInstances<DictTest>(d, d2);

    PrimitiveTest p;
    PrimitiveTest p2;
    runTestWithInstances<PrimitiveTest>(p, p2);

    SimpleClass o1;
    SimpleClass o12;
    runTestWithInstances<SimpleClass>(o1, o12);

    ObjectTest o2;
    ObjectTest o22;
    runTestWithInstances<ObjectTest>(o2, o22);

    ImageTest i;
    ImageTest i2;
    runTestWithInstances<ImageTest>(i, i2);
}
