/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// STD/STL
#include <string>
#include <cctype>
#include <map>

// Boost
#include <boost/algorithm/string.hpp>

// Header
#include "AronAbstractTypeNavigator.h"

// ArmarX
#include <RobotAPI/libraries/aron/types/AronAbstractTypeNavigatorFactory.h>

namespace armarx
{
    namespace aron
    {
        namespace types
        {
            // constantes
            const std::string AronAbstractTypeNavigator::ARON_DATA_NAME = boost::core::demangle(typeid(AronData).name());
            const std::string AronAbstractTypeNavigator::ARON_DATA_PTR_NAME = AronAbstractTypeNavigator::ARON_DATA_NAME + "::PointerType";

            const std::string AronAbstractTypeNavigator::ARON_TYPE_NAME = boost::core::demangle(typeid(AronAbstractType).name());
            const std::string AronAbstractTypeNavigator::ARON_TYPE_PTR_NAME = AronAbstractTypeNavigator::ARON_TYPE_NAME + "::PointerType";

            std::map<std::string, AronObjectTypeNavigatorPtr> AronAbstractTypeNavigator::AllAronObjectTypes = {};


            namespace
            {
                // anonymous namespace
                std::string resolveCppTypeName(const std::string& realTypeName)
                {
                    if (realTypeName == boost::core::demangle(typeid(std::string).name()))
                    {
                        return "std::string";
                    }
                    return realTypeName;
                }
            }

            // constructors
            AronAbstractTypeNavigator::AronAbstractTypeNavigator(const AronTypeDescriptor& d, const std::string& cpp, const std::string& aron, const std::string& type) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(d),
                cppTypename(resolveCppTypeName(cpp)),
                aronTypename(aron),
                aronTypeTypename(type)
            {

            }

            bool AronAbstractTypeNavigator::operator<(const AronAbstractTypeNavigator& other) const
            {
                // Typenames must be unique
                return getFullCppTypename() < other.getFullCppTypename();
            }

            std::string AronAbstractTypeNavigator::getFullCppTypename() const
            {
                return cppTypename;
            }

            std::string AronAbstractTypeNavigator::getCppTypename() const
            {
                std::vector<std::string> split;
                std::string name = getFullCppTypename();
                boost::algorithm::split(split, name, boost::is_any_of("::"));
                std::string type = split[split.size() - 1];
                type[0] = toupper(type[0]);
                return type;
            }

            std::string AronAbstractTypeNavigator::getFullAronTypename() const
            {
                return aronTypename;
            }

            std::string AronAbstractTypeNavigator::getAronTypename() const
            {
                std::vector<std::string> split;
                std::string name = getFullAronTypename();
                boost::algorithm::split(split, name, boost::is_any_of("::"));
                std::string type = split[split.size() - 1];
                type[0] = toupper(type[0]);
                return type;
            }

            std::string AronAbstractTypeNavigator::getFullAronPtrTypename() const
            {
                return getFullAronTypename() + "::PointerType";
            }

            std::string AronAbstractTypeNavigator::getFullAronTypeTypename() const
            {
                return aronTypeTypename;
            }

            std::string AronAbstractTypeNavigator::getAronTypeTypename() const
            {
                std::vector<std::string> split;
                std::string name = getFullAronTypeTypename();
                boost::algorithm::split(split, name, boost::is_any_of("::"));
                std::string type = split[split.size() - 1];
                type[0] = toupper(type[0]);
                return type;
            }

            std::string AronAbstractTypeNavigator::getFullAronTypePtrTypename() const
            {
                return getFullAronTypeTypename() + "::PointerType";
            }

            std::string AronAbstractTypeNavigator::escapeAccessor(const std::string& accessor)
            {
                std::string escaped_accessor = accessor;
                boost::algorithm::replace_all(escaped_accessor, "->", "__");
                boost::algorithm::replace_all(escaped_accessor, ".", "__");
                boost::algorithm::replace_all(escaped_accessor, "[", "___");
                boost::algorithm::replace_all(escaped_accessor, "]", "___");
                return escaped_accessor;
            }

            CppMethodPtr AronAbstractTypeNavigator::toResetMethod() const
            {
                std::stringstream doc;
                doc << "@brief reset() - This method resets all member variables to default. \n";
                doc << "                 Please note that this method does not change the internal structure, i.e. images get \n";
                doc << "                 resetted BUT have the same dimension as before. \n";
                doc << "                 If you want to re-initialize the whole container, please use 'initialize()' instzead. \n";
                doc << "@return - nothing";

                CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void reset() override", doc.str()));
                CppBlockPtr b = this->getResetBlock("");
                m->setBlock(b);
                return m;
            }

            CppMethodPtr AronAbstractTypeNavigator::toInitializeMethod() const
            {
                std::stringstream doc;
                doc << "@brief initialize() - This method initializeses class members to default and sets special base-class variables. \n";
                doc << "                      Currently, this method initializes the member variables as well as the aron type variable.\n";
                doc << "@return - nothing";

                CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void initialize() override", doc.str()));
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("///////////////////////////////////////////////////////////////////////////////");
                b->addLine("//INITIALIZE MEMBER VARIABLES//////////////////////////////////////////////////");
                b->addLine("///////////////////////////////////////////////////////////////////////////////");
                CppBlockPtr b2 = this->getInitializeMembersBlock("");
                b = CppBlock::mergeBlocks(b, b2);

                b->addLine("///////////////////////////////////////////////////////////////////////////////");
                b->addLine("//INITIALIZE ARON TYPE VARIABLE////////////////////////////////////////////////");
                b->addLine("///////////////////////////////////////////////////////////////////////////////");
                CppBlockPtr b3 = this->getInitializeAronTypeBlock("aronTypeRoot");
                b3->addLine("aronType = aronTypeRoot;");
                b = CppBlock::mergeBlocks(b, b3);

                m->setBlock(b);
                return m;
            }

            CppMethodPtr AronAbstractTypeNavigator::toWriteMethod() const
            {
                std::stringstream doc;
                doc << "@brief write() - This method returns a new type from the member data types using a writer implementation. \n";
                doc << "@param w - The writer implementation\n";
                doc << "@return - the result of the writer implementation";

                CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void write(armarx::aron::io::AronWriter& w) const override", doc.str()));
                CppBlockPtr b = this->getWriteBlock("");
                m->setBlock(b);
                return m;
            }

            CppMethodPtr AronAbstractTypeNavigator::toReadMethod() const
            {
                std::stringstream doc;
                doc << "@brief read() - This method sets the struct members to new values given in a reader implementation. \n";
                doc << "@param r - The reader implementation\n";
                doc << "@return - nothing";

                CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void read(armarx::aron::io::AronReader& r) override", doc.str()));
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("reset();");
                b->addLine("");
                CppBlockPtr b2 = this->getReadBlock("");
                b = CppBlock::mergeBlocks(b, b2);
                m->setBlock(b);
                return m;
            }

            CppMethodPtr AronAbstractTypeNavigator::toSpecializedWriterMethod(const std::string& returnname, const std::string& methodname, const std::string& writerName) const
            {
                std::stringstream doc;
                doc << "@brief specializedWrite() - This method returns a new type from the member data types using a writer implementation. \n";
                doc << "@return - the result of the writer implementation";

                CppMethodPtr m = CppMethodPtr(new CppMethod(returnname + " " + methodname + "() const", doc.str()));
                m->addLine(writerName + " writer;");
                m->addLine("this->write(writer);");
                m->addLine("return writer.getResult();");
                return m;
            }

            CppMethodPtr AronAbstractTypeNavigator::toSpecializedReaderMethod(const std::string& argumentname, const std::string& methodname, const std::string& readerName) const
            {
                std::stringstream doc;
                doc << "@brief specializedRead() - This method sets the struct members to new values given in a reader implementation. \n";
                doc << "@return - nothing";

                CppMethodPtr m = CppMethodPtr(new CppMethod("void " + methodname + "(const " + argumentname + "& input)", doc.str()));
                m->addLine(readerName + " reader(input);");
                m->addLine("this->read(reader);");
                return m;
            }

            CppMethodPtr AronAbstractTypeNavigator::toEqualsMethod() const
            {
                std::stringstream doc;
                doc << "@brief operator==() - This method checks whether all values equal another instance. \n";
                doc << "@param i - The other instance\n";
                doc << "@return - true, if all members are the same, false otherwise";

                CppMethodPtr m = CppMethodPtr(new CppMethod("bool operator==(const " + this->getFullCppTypename() + "& i) const", doc.str()));
                CppBlockPtr b = this->getEqualsBlock("", "i");
                b->addLine("return true;");
                m->setBlock(b);
                return m;
            }

            // static methods
            AronAbstractTypeNavigatorPtr AronAbstractTypeNavigator::GetAronAbstractTypePtrByName(const std::string& name)
            {
                const std::string lower_name = boost::algorithm::to_lower_copy(name);
                auto it = AllAronObjectTypes.find(lower_name);
                if (it == AllAronObjectTypes.end())
                {
                    throw LocalException("AronAbstractTypeNavigator: Could not find a valid object type for name '" + lower_name + "', (original: " + name + ").");
                }
                return it->second;
            }

            void AronAbstractTypeNavigator::AddObjectTypeToAllObjectTypesList(const AronObjectTypeNavigatorPtr& n)
            {
                checkAbstractTypeNavigatorPtrForNull(n, "");
                if (AllAronObjectTypes.find(boost::algorithm::to_lower_copy(n->getFullCppTypename())) != AllAronObjectTypes.end())
                {
                    throw LocalException("AronAbstractTypeNavigator: The object type '" + n->getFullCppTypename() + "' has already beed added to the AllAronAbstractTypes.");
                }
                AllAronObjectTypes[boost::algorithm::to_lower_copy(n->getFullCppTypename())] = n;
            }

            void AronAbstractTypeNavigator::checkAbstractTypeNavigatorPtrForNull(const AronAbstractTypeNavigatorPtr& data, const std::string& message)
            {
                if (data.get() == nullptr)
                {
                    throw LocalException("AronAbstractTypeNavigator: Could not cast an AronAbstractTypeNavigatorPtr. The Ptr was NULL. " + message);
                }
            }
        }
    }
}


