/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <memory>
#include <vector>
#include <map>

// Boost
#include <boost/core/demangle.hpp>
#include <boost/algorithm/string.hpp>

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>
#include <ArmarXCore/libraries/cppgen/CppMethod.h>
#include <ArmarXCore/libraries/cppgen/CppClass.h>

#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/AronNavigator.h>

namespace armarx
{
    namespace aron
    {
        namespace types
        {
            class AronAbstractTypeNavigatorFactory;
            typedef std::shared_ptr<AronAbstractTypeNavigatorFactory> AronAbstractTypeNavigatorFactoryPtr;

            class AronObjectTypeNavigator;
            typedef std::shared_ptr<AronObjectTypeNavigator> AronObjectTypeNavigatorPtr;

            class AronAbstractTypeNavigator;
            typedef std::shared_ptr<AronAbstractTypeNavigator> AronAbstractTypeNavigatorPtr;

            class AronAbstractTypeNavigator :
                virtual public AronNavigator<AronTypeDescriptor, AronAbstractType>
            {
            public:
                using PointerType = AronAbstractTypeNavigatorPtr;

            public:
                // constructors
                AronAbstractTypeNavigator(const AronTypeDescriptor&, const std::string&, const std::string&, const std::string&);

                // operators
                bool operator<(const AronAbstractTypeNavigator&) const;

                // public member methods
                std::string getFullCppTypename() const;
                std::string getCppTypename() const;

                std::string getFullAronTypename() const;
                std::string getAronTypename() const;
                std::string getFullAronPtrTypename() const;

                std::string getFullAronTypeTypename() const;
                std::string getAronTypeTypename() const;
                std::string getFullAronTypePtrTypename() const;

                CppMethodPtr toSpecializedWriterMethod(const std::string&, const std::string&, const std::string&) const;
                CppMethodPtr toSpecializedReaderMethod(const std::string&, const std::string&, const std::string&) const;


                // virtual override definitions
                CppMethodPtr toResetMethod() const;
                virtual CppBlockPtr getResetBlock(const std::string&) const = 0;

                CppMethodPtr toInitializeMethod() const;
                virtual CppBlockPtr getInitializeAronTypeBlock(const std::string&) const = 0;
                virtual CppBlockPtr getInitializeMembersBlock(const std::string&) const = 0;

                CppMethodPtr toWriteMethod() const;
                virtual CppBlockPtr getWriteBlock(const std::string&) const = 0;

                CppMethodPtr toReadMethod() const;
                virtual CppBlockPtr getReadBlock(const std::string&) const = 0;

                CppMethodPtr toEqualsMethod() const;
                virtual CppBlockPtr getEqualsBlock(const std::string&, const std::string&) const = 0;

                virtual AronAbstractTypePtr getResult() const override = 0;

                // static methods
                static AronAbstractTypeNavigatorPtr GetAronAbstractTypePtrByName(const std::string& name);
                static void AddObjectTypeToAllObjectTypesList(const AronObjectTypeNavigatorPtr&);

                template<typename AronAbstractTypeNavigatorClass>
                static typename AronAbstractTypeNavigatorClass::PointerType DynamicCast(const AronAbstractTypeNavigatorPtr& n)
                {
                    return AronAbstractTypeNavigatorClass::DynamicCast(n);
                }

            protected:
                static void checkAbstractTypeNavigatorPtrForNull(const AronAbstractTypeNavigatorPtr& data, const std::string& message);
                static std::string escapeAccessor(const std::string&);

            public:
                // public constantes
                const static std::string ARON_DATA_NAME;
                const static std::string ARON_DATA_PTR_NAME;

                const static std::string ARON_TYPE_NAME;
                const static std::string ARON_TYPE_PTR_NAME;

            private:
                // members
                std::string cppTypename;
                std::string aronTypename;
                std::string aronTypeTypename;

                // static members
                static std::map<std::string, AronObjectTypeNavigatorPtr> AllAronObjectTypes;
            };
        }
    }
}
