/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Boost
#include <boost/algorithm/string.hpp>

// Header
#include "AronDictTypeNavigator.h"


namespace armarx
{
    namespace aron
    {
        namespace types
        {
            // constructors
            AronDictTypeNavigator::AronDictTypeNavigator(const std::string& acceptedTypename, const AronNavigatorPath& path) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(AronTypeDescriptor::eAronDictType),
                AronAbstractTypeNavigator(AronTypeDescriptor::eAronDictType, "std::map<std::string, " + acceptedTypename + ">", boost::core::demangle(typeid(AronDict).name()), boost::core::demangle(typeid(AronDictType).name())),
                acceptedTypeNavigator(GetAronAbstractTypePtrByName(acceptedTypename)),
                type(new AronDictType())
            {
                type->acceptedType = acceptedTypeNavigator->getResult();
            }

            AronDictTypeNavigator::AronDictTypeNavigator(const AronAbstractTypeNavigatorPtr& acceptedType, const AronNavigatorPath& path) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(AronTypeDescriptor::eAronDictType),
                AronAbstractTypeNavigator(AronTypeDescriptor::eAronDictType, "std::map<std::string, " + acceptedType->getFullCppTypename() + ">", boost::core::demangle(typeid(AronDict).name()), boost::core::demangle(typeid(AronDictType).name())),
                acceptedTypeNavigator(acceptedType),
                type(new AronDictType())
            {
                type->acceptedType = acceptedTypeNavigator->getResult();
            }

            // static methods
            AronDictTypeNavigatorPtr AronDictTypeNavigator::DynamicCast(const AronAbstractTypeNavigatorPtr& n)
            {
                AronDictTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronDictTypeNavigator>(n);
                checkAbstractTypeNavigatorPtrForNull(casted, "");
                return casted;
            }

            // virtual implementations
            AronAbstractTypePtr AronDictTypeNavigator::getResult() const
            {
                return type;
            }

            CppBlockPtr AronDictTypeNavigator::getResetBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(accessor + ".clear();");
                return b;
            }

            CppBlockPtr AronDictTypeNavigator::getInitializeMembersBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(accessor + " = {};");
                return b;
            }


            CppBlockPtr AronDictTypeNavigator::getInitializeAronTypeBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());

                std::string escaped_accessor = escapeAccessor(accessor);
                std::string accessor_iterator = escaped_accessor + "_dicttype";

                CppBlockPtr b2 = acceptedTypeNavigator->getInitializeAronTypeBlock(accessor_iterator);
                b = CppBlock::mergeBlocks(b, b2);

                b->addLine(getFullAronTypePtrTypename() + " " + accessor + " = " + getFullAronTypePtrTypename() + "(new " + getFullAronTypeTypename() + "(" + accessor_iterator + "));");
                return b;
            }

            CppBlockPtr AronDictTypeNavigator::getWriteBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("w.writeStartDict();");

                std::string escaped_accessor = escapeAccessor(accessor);
                std::string accessor_iterator_key = escaped_accessor + "_dictkey";
                std::string accessor_iterator_val = escaped_accessor + "_dictval";

                b->addLine("for (const auto& [" + accessor_iterator_key + ", " + accessor_iterator_val + "] : " + accessor + ") ");

                CppBlockPtr b2 = CppBlockPtr(new CppBlock());
                b2->addLine("w.writeKey(" + accessor_iterator_key + ");");
                b2 = CppBlock::mergeBlocks(b2, acceptedTypeNavigator->getWriteBlock(accessor_iterator_val));
                b->addBlock(b2);

                b->addLine("w.writeEndDict();");
                return b;
            }

            CppBlockPtr AronDictTypeNavigator::getReadBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("r.readStartDict();");

                std::string escaped_accessor = escapeAccessor(accessor);
                std::string accessor_iterator = escaped_accessor + "_dictiterator";
                std::string accessor_iterator_key = escaped_accessor + "_dictkey";

                b->addLine("while(!r.readEndDict())");
                CppBlockPtr b2 = CppBlockPtr(new CppBlock());
                b2->addLine("std::string " + accessor_iterator_key + " = r.readKey();");
                b2->addLine(acceptedTypeNavigator->getFullCppTypename() + " " + accessor_iterator + ";");
                b2 = CppBlock::mergeBlocks(b2, acceptedTypeNavigator->getReadBlock(accessor_iterator));
                b2->addLine(accessor + "[" + accessor_iterator_key + "] = " + accessor_iterator + ";");

                b->addBlock(b2);
                return b;
            }

            CppBlockPtr AronDictTypeNavigator::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("if (not (" + accessor + " == " + otherInstanceAccessor + "))");
                b->addLine("\t return false;");
                return b;
            }
        }
    }
}

