/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Boost
#include <boost/algorithm/string.hpp>

// Header
#include "AronEigenMatrixTypeNavigator.h"

namespace armarx
{
    namespace aron
    {
        namespace types
        {

            const std::set<std::string> AronEigenMatrixTypeNavigator::acceptedTypes =
            {
                "short",
                "int",
                "long",
                "float",
                "double",
                //"std::complex<float>",
                //"std::complex<double>"
            };

            namespace
            {
                // anonymous helper functions
                std::string implodeDimensionVector(const std::vector<int>& dimensions)
                {
                    if (dimensions.size() == 0)
                    {
                        return "";
                    }

                    std::stringstream ss;
                    for (unsigned int i = 0; i < dimensions.size() - 1; ++i)
                    {
                        ss << dimensions[i] << ", ";
                    }
                    ss << dimensions[dimensions.size() - 1];
                    return ss.str();
                }
            }

            // constructors
            AronEigenMatrixTypeNavigator::AronEigenMatrixTypeNavigator(const std::vector<int>& dims, const std::string& t, const AronNavigatorPath& path) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(AronTypeDescriptor::eAronEigenMatrixType),
                AronAbstractTypeNavigator(AronTypeDescriptor::eAronEigenMatrixType, "::Eigen::Matrix<" + t + ", " + implodeDimensionVector(dims) + ">", boost::core::demangle(typeid(AronBlob).name()), boost::core::demangle(typeid(AronEigenMatrixType).name())),
                dimensions(dims),
                type(new AronEigenMatrixType())
            {
                type->dimensions = dims;
                type->typeName = t;

                bool found = false;
                for (const auto& typeName : acceptedTypes)
                {
                    if (typeName == t)
                    {
                        this->usedType = typeName;
                        found = true;
                        break;
                    }
                }

                if (!found)
                {
                    throw LocalException("AronEigenMatrixTypeNavigator: Could not find the typename '" + t + "' in accepted types.");
                }
            }

            // static methods
            AronEigenMatrixTypeNavigatorPtr AronEigenMatrixTypeNavigator::DynamicCast(const AronAbstractTypeNavigatorPtr& n)
            {
                AronEigenMatrixTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronEigenMatrixTypeNavigator>(n);
                checkAbstractTypeNavigatorPtrForNull(casted, "");
                return casted;
            }

            // virtual implementations
            AronAbstractTypePtr AronEigenMatrixTypeNavigator::getResult() const
            {
                return type;
            }

            CppBlockPtr AronEigenMatrixTypeNavigator::getResetBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(accessor + " = {};");
                return b;
            }

            CppBlockPtr AronEigenMatrixTypeNavigator::getInitializeAronTypeBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(getFullAronTypePtrTypename() + " " + accessor + " = " + getFullAronTypePtrTypename() + "(new " + getFullAronTypeTypename() + "());");
                return b;
            }

            CppBlockPtr AronEigenMatrixTypeNavigator::getInitializeMembersBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(accessor + " = {};");
                return b;
            }

            CppBlockPtr AronEigenMatrixTypeNavigator::getWriteBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("w.writeBlob(reinterpret_cast<unsigned char*>(" + accessor + ".data()));");
                return b;
            }

            CppBlockPtr AronEigenMatrixTypeNavigator::getReadBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(accessor + " = " + getFullCppTypename() + "(reinterpret_cast<" + usedType + "*>(r.readBlob())); // Eigen already performs memcpy");
                return b;
            }

            CppBlockPtr AronEigenMatrixTypeNavigator::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                return b;
            }
        }
    }
}

