/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Boost
#include <boost/algorithm/string.hpp>

// Header
#include "AronIVTCByteImageTypeNavigator.h"

namespace armarx
{
    namespace aron
    {
        namespace types
        {

            const std::map<std::string, unsigned int> AronIVTCByteImageTypeNavigator::acceptedTypes =
            {
                {"CByteImage::ImageType::eGrayScale", 1},
                {"CByteImage::ImageType::eRGB24", 3},
                {"CByteImage::ImageType::eRGB24Split", 3}
            };

            // constructors
            AronIVTCByteImageTypeNavigator::AronIVTCByteImageTypeNavigator(const unsigned int w, const unsigned int h, const std::string& t, const AronNavigatorPath& path) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(AronTypeDescriptor::eAronIVTCByteImageType),
                AronAbstractTypeNavigator(AronTypeDescriptor::eAronIVTCByteImageType, "std::shared_ptr<CByteImage>", boost::core::demangle(typeid(AronBlob).name()), boost::core::demangle(typeid(AronIVTCByteImageType).name())),
                width(w),
                height(h),
                type(new AronIVTCByteImageType())
            {
                type->width = w;
                type->height = h;
                type->typeName = t;

                auto i = acceptedTypes.find(t);
                if (acceptedTypes.find(t) == acceptedTypes.end())
                {
                    throw LocalException("AronIVTCByteImageTypeNavigator: Could not find the typename '" + t + "' in accepted types.");
                }
                else
                {
                    this->usedType = i->first;
                    this->elementSize = i->second;
                }
            }

            // static methods
            AronIVTCByteImageTypeNavigatorPtr AronIVTCByteImageTypeNavigator::DynamicCast(const AronAbstractTypeNavigatorPtr& n)
            {
                AronIVTCByteImageTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronIVTCByteImageTypeNavigator>(n);
                checkAbstractTypeNavigatorPtrForNull(casted, "");
                return casted;
            }

            // virtual implementations
            AronAbstractTypePtr AronIVTCByteImageTypeNavigator::getResult() const
            {
                return type;
            }

            CppBlockPtr AronIVTCByteImageTypeNavigator::getResetBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(accessor + " = " + getFullCppTypename() + "(new CByteImage(" + accessor + "->width, " + accessor + "->height, " + accessor + "->type));");
                return b;
            }

            CppBlockPtr AronIVTCByteImageTypeNavigator::getInitializeMembersBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(accessor + " = " + getFullCppTypename() + "(new CByteImage(" + std::to_string(this->width) + ", " + std::to_string(this->height) + ", " + this->usedType + "));");
                return b;
            }

            CppBlockPtr AronIVTCByteImageTypeNavigator::getInitializeAronTypeBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(getFullAronTypePtrTypename() + " " + accessor + " = " + getFullAronTypePtrTypename() + "(new " + getFullAronTypeTypename() + "());");
                return b;
            }

            CppBlockPtr AronIVTCByteImageTypeNavigator::getWriteBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("w.writeBlob(" + accessor + "->width * " + accessor + "->height * " + accessor + "->bytesPerPixel, reinterpret_cast<unsigned char*>(" + accessor + "->pixels));");
                return b;
            }

            CppBlockPtr AronIVTCByteImageTypeNavigator::getReadBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(accessor + " = " + getFullCppTypename() + "(new CByteImage(" + accessor + "->width, " + accessor + "->height, " + accessor + "->type, false));");
                b->addLine("memcpy(" + accessor + "->pixels, reinterpret_cast<unsigned char*>(r.readBlob()), " + accessor + "->width * " + accessor + "->height * " + accessor + "->bytesPerPixel" + ");");
                return b;
            }

            CppBlockPtr AronIVTCByteImageTypeNavigator::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("if (not (" + accessor + "->IsCompatible(" + otherInstanceAccessor + ".get())))");
                b->addLine("\t return false;");
                b->addLine("if (not (memcmp(" + accessor + "->pixels, " + otherInstanceAccessor + "->pixels, " + accessor + "->width * " + accessor + "->height * " + accessor + "->bytesPerPixel) == 0))");
                b->addLine("\t return false;");
                return b;
            }
        }
    }
}

