/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Boost
#include <boost/algorithm/string/replace.hpp>

// Header
#include "AronListTypeNavigator.h"

namespace armarx
{
    namespace aron
    {
        namespace types
        {
            // constructors
            AronListTypeNavigator::AronListTypeNavigator(const AronAbstractTypeNavigatorPtr& acceptedType, const AronNavigatorPath& path) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(AronTypeDescriptor::eAronListType),
                AronAbstractTypeNavigator(AronTypeDescriptor::eAronListType, "std::vector<" + acceptedType->getFullCppTypename() + ">", boost::core::demangle(typeid(AronList).name()), boost::core::demangle(typeid(AronListType).name())),
                acceptedTypeNavigator(acceptedType),
                type(new AronListType())
            {
                type->acceptedType = acceptedTypeNavigator->getResult();
            }

            AronListTypeNavigator::AronListTypeNavigator(const std::string& acceptedTypename, const AronNavigatorPath& path) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(AronTypeDescriptor::eAronListType),
                AronAbstractTypeNavigator(AronTypeDescriptor::eAronListType, "std::vector<" + acceptedTypename + ">", boost::core::demangle(typeid(AronList).name()), boost::core::demangle(typeid(AronListType).name())),
                acceptedTypeNavigator(GetAronAbstractTypePtrByName(acceptedTypename)),
                type(new AronListType())
            {
                type->acceptedType = acceptedTypeNavigator->getResult();
            }

            // static methods
            AronListTypeNavigatorPtr AronListTypeNavigator::DynamicCast(const AronAbstractTypeNavigatorPtr& n)
            {
                AronListTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronListTypeNavigator>(n);
                checkAbstractTypeNavigatorPtrForNull(casted, "");
                return casted;
            }

            // virtual implementations
            AronAbstractTypePtr AronListTypeNavigator::getResult() const
            {
                return type;
            }

            CppBlockPtr AronListTypeNavigator::getResetBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(accessor + ".clear();");
                return b;
            }

            CppBlockPtr AronListTypeNavigator::getInitializeMembersBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine(accessor + " = {};");
                return b;
            }

            CppBlockPtr AronListTypeNavigator::getInitializeAronTypeBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());

                std::string escaped_accessor = escapeAccessor(accessor);
                std::string accessor_iterator = escaped_accessor + "_listtype";

                CppBlockPtr b2 = acceptedTypeNavigator->getInitializeAronTypeBlock(accessor_iterator);
                b = CppBlock::mergeBlocks(b, b2);

                b->addLine(getFullAronTypePtrTypename() + " " + accessor + " = " + getFullAronTypePtrTypename() + "(new " + getFullAronTypeTypename() + "(" + accessor_iterator + "));");
                return b;
            }

            CppBlockPtr AronListTypeNavigator::getWriteBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("w.writeStartList();");

                std::string escaped_accessor = escapeAccessor(accessor);
                std::string accessor_iterator = escaped_accessor + "_listiterator";

                b->addLine("for(unsigned int " + accessor_iterator + " = 0; " + accessor_iterator + " < " + accessor + ".size(); ++" + accessor_iterator + ")");
                CppBlockPtr b2 = acceptedTypeNavigator->getWriteBlock(accessor + "[" + accessor_iterator + "]");
                b->addBlock(b2);
                b->addLine("w.writeEndList();");
                return b;
            }

            CppBlockPtr AronListTypeNavigator::getReadBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("r.readStartList();");

                std::string escaped_accessor = escapeAccessor(accessor);
                std::string accessor_iterator = escaped_accessor + "_listiterator";

                b->addLine("while(!r.readEndList())");
                CppBlockPtr b2 = CppBlockPtr(new CppBlock());
                b2->addLine(acceptedTypeNavigator->getFullCppTypename() + " " + accessor_iterator + ";");
                b2 = CppBlock::mergeBlocks(b2, acceptedTypeNavigator->getReadBlock(accessor_iterator));
                b2->addLine(accessor + ".push_back(" + accessor_iterator + ");");

                b->addBlock(b2);
                return b;
            }

            CppBlockPtr AronListTypeNavigator::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("if (not (" + accessor + " == " + otherInstanceAccessor + "))");
                b->addLine("\t return false;");
                return b;
            }
        }
    }
}
