/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Boost
#include <boost/algorithm/string.hpp>

// Header
#include "AronObjectTypeNavigator.h"


namespace armarx
{
    namespace aron
    {
        namespace types
        {

            // constructors
            AronObjectTypeNavigator::AronObjectTypeNavigator(const std::string& cppName, const std::map<std::string, AronAbstractTypeNavigatorPtr>& children, const AronNavigatorPath& path) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(AronTypeDescriptor::eAronObjectType),
                AronAbstractTypeNavigator(AronTypeDescriptor::eAronObjectType, cppName, boost::core::demangle(typeid(AronDict).name()), boost::core::demangle(typeid(AronObjectType).name())),
                childrenNavigators(children),
                type(new AronObjectType())
            {
                AddObjectTypeToAllObjectTypesList(AronObjectTypeNavigatorPtr(this));

                for (const auto& [key, nav] : childrenNavigators)
                {
                    type->elementTypes[key] = nav->getResult();
                }
            }

            AronObjectTypeNavigator::AronObjectTypeNavigator(const std::string& cppName, const AronNavigatorPath& path) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(AronTypeDescriptor::eAronObjectType),
                AronAbstractTypeNavigator(AronTypeDescriptor::eAronObjectType, cppName, boost::core::demangle(typeid(AronDict).name()), boost::core::demangle(typeid(AronObjectType).name())),
                type(new AronObjectType())
            {
                AddObjectTypeToAllObjectTypesList(AronObjectTypeNavigatorPtr(this));

                type->name = getFullCppTypename();
            }

            // static methods
            AronObjectTypeNavigatorPtr AronObjectTypeNavigator::DynamicCast(const AronAbstractTypeNavigatorPtr& n)
            {
                AronObjectTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronObjectTypeNavigator>(n);
                checkAbstractTypeNavigatorPtrForNull(casted, "");
                return casted;
            }

            // public member functions
            void AronObjectTypeNavigator::addElement(const std::string& k, const AronAbstractTypeNavigatorPtr& v)
            {
                childrenNavigators[k] = v;
                type->elementTypes[k] = v->getResult();
            }

            std::map<std::string, AronAbstractTypeNavigatorPtr> AronObjectTypeNavigator::getChildren() const
            {
                return childrenNavigators;
            }

            // virtual implementations
            AronAbstractTypePtr AronObjectTypeNavigator::getResult() const
            {
                return type;
            }

            CppBlockPtr AronObjectTypeNavigator::getResetBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                for (const auto& [key, child] : childrenNavigators)
                {
                    CppBlockPtr b2 = child->getResetBlock((accessor != "" ? (accessor + "." + key) : key));
                    b = CppBlock::mergeBlocks(b, b2);
                }
                return b;
            }

            CppBlockPtr AronObjectTypeNavigator::getInitializeMembersBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                for (const auto& [key, child] : childrenNavigators)
                {
                    CppBlockPtr b2 = child->getInitializeMembersBlock((accessor != "" ? (accessor + "." + key) : key));
                    b = CppBlock::mergeBlocks(b, b2);
                }
                return b;
            }

            CppBlockPtr AronObjectTypeNavigator::getInitializeAronTypeBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());

                std::string escaped_accessor = escapeAccessor(accessor);
                std::string accessor_iterator_container = escaped_accessor + "_objectiteratorcontainer";

                b->addLine("std::map<std::string, " + ARON_TYPE_PTR_NAME + "> " + accessor_iterator_container + ";");

                for (const auto& [key, child] : childrenNavigators)
                {
                    std::string accessor_iterator = escaped_accessor + "_" + key;
                    CppBlockPtr b2 = child->getInitializeAronTypeBlock(accessor_iterator);
                    b = CppBlock::mergeBlocks(b, b2);

                    b->addLine(accessor_iterator_container + "[\"" + key + "\"] = " + accessor_iterator + ";");
                }
                b->addLine(getFullAronTypePtrTypename() + " " + accessor + " = " + getFullAronTypePtrTypename() + "(new " + getFullAronTypeTypename() + "(\"" + getFullCppTypename() + "\", " + accessor_iterator_container + "));");
                return b;
            }

            CppBlockPtr AronObjectTypeNavigator::getWriteBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("w.writeStartDict();");
                for (const auto& [key, child] : childrenNavigators)
                {
                    b->addLine("w.writeKey(\"" + key + "\");");
                    CppBlockPtr b2 = child->getWriteBlock((accessor != "" ? (accessor + "." + key) : key));
                    b = CppBlock::mergeBlocks(b, b2);
                }
                b->addLine("w.writeEndDict();");
                return b;
            }

            CppBlockPtr AronObjectTypeNavigator::getReadBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("r.readStartDict();");
                for (const auto& [key, child] : childrenNavigators)
                {
                    b->addLine("r.readMember(\"" + key + "\");");
                    CppBlockPtr b2 = child->getReadBlock((accessor != "" ? (accessor + "." + key) : key));
                    b = CppBlock::mergeBlocks(b, b2);
                }
                b->addLine("r.readEndDict();");
                return b;
            }

            CppBlockPtr AronObjectTypeNavigator::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                if (accessor == "") // object defintiion
                {
                    for (const auto& [key, child] : this->getChildren())
                    {
                        CppBlockPtr b2 = child->getEqualsBlock(key, otherInstanceAccessor + "." + key);
                        b = CppBlock::mergeBlocks(b, b2);
                    }
                }
                else // object usage with existing operator==
                {
                    b->addLine("if (not (" + accessor + " == " + otherInstanceAccessor + "))");
                    b->addLine("\t return false;");
                }
                return b;
            }
        }
    }
}

