/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Boost
#include <boost/algorithm/string.hpp>

// Header
#include "AronTupleTypeNavigator.h"


namespace armarx
{
    namespace aron
    {
        namespace types
        {
            namespace
            {
                // anonymous helper functions
                std::vector<std::string> getAronTupleTypeNames(const std::vector<AronAbstractTypeNavigatorPtr>& arons)
                {
                    std::vector<std::string> names;
                    for (const auto& aron : arons)
                    {
                        names.push_back(aron->getFullCppTypename());
                    }
                    return names;
                }
            }

            // constructors
            AronTupleTypeNavigator::AronTupleTypeNavigator(const std::vector<std::string>& acceptedCppTypenames, const AronNavigatorPath& path) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(AronTypeDescriptor::eAronTupleType),
                AronAbstractTypeNavigator(AronTypeDescriptor::eAronTupleType, "std::tuple<" + boost::algorithm::join(acceptedCppTypenames, ", ") + ">", boost::core::demangle(typeid(AronList).name()), boost::core::demangle(typeid(AronTupleType).name())),
                type(new AronTupleType())
            {
                for (const std::string& typeName : acceptedCppTypenames)
                {
                    AronAbstractTypeNavigatorPtr nav = GetAronAbstractTypePtrByName(typeName);
                    acceptedTypeNavigators.push_back(nav);
                    type->elementTypes.push_back(nav->getResult());
                }
            }

            AronTupleTypeNavigator::AronTupleTypeNavigator(const std::vector<AronAbstractTypeNavigatorPtr>& acceptedTypeNavigators, const AronNavigatorPath& path) :
                AronNavigator<AronTypeDescriptor, AronAbstractType>::AronNavigator(AronTypeDescriptor::eAronTupleType),
                AronAbstractTypeNavigator(AronTypeDescriptor::eAronTupleType, "std::tuple<" + boost::algorithm::join(getAronTupleTypeNames(acceptedTypeNavigators), ", ") + ">", boost::core::demangle(typeid(AronList).name()), boost::core::demangle(typeid(AronTupleType).name())),
                acceptedTypeNavigators(acceptedTypeNavigators),
                type(new AronTupleType())
            {

            }

            // static methods
            AronTupleTypeNavigatorPtr AronTupleTypeNavigator::DynamicCast(const AronAbstractTypeNavigatorPtr& n)
            {
                AronTupleTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronTupleTypeNavigator>(n);
                checkAbstractTypeNavigatorPtrForNull(casted, "");
                return casted;
            }

            // public member functions
            void AronTupleTypeNavigator::addElement(const AronAbstractTypeNavigatorPtr& v)
            {
                acceptedTypeNavigators.push_back(v);
                type->elementTypes.push_back(v->getResult());
            }

            std::vector<AronAbstractTypeNavigatorPtr> AronTupleTypeNavigator::getElements() const
            {
                return acceptedTypeNavigators;
            }

            // virtual implementations
            AronAbstractTypePtr AronTupleTypeNavigator::getResult() const
            {
                return type;
            }

            CppBlockPtr AronTupleTypeNavigator::getInitializeMembersBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());

                unsigned int i = 0;
                for (const auto& child : acceptedTypeNavigators)
                {
                    CppBlockPtr b2 = child->getInitializeMembersBlock("std::get<" + std::to_string(i++) + ">(" + accessor + ")");
                    b = CppBlock::mergeBlocks(b, b2);
                }
                return b;
            }

            CppBlockPtr AronTupleTypeNavigator::getInitializeAronTypeBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());

                std::string escaped_accessor = escapeAccessor(accessor);
                std::string accessor_iterator_container = escaped_accessor + "_tupleiteratorcontainer";

                b->addLine("std::vector<" + getFullAronTypePtrTypename() + "> " + accessor_iterator_container + ";");

                unsigned int i = 0;
                for (const auto& child : acceptedTypeNavigators)
                {
                    std::string accessor_iterator = escaped_accessor + "_tupleiterator" + std::to_string(i++);
                    CppBlockPtr b2 = child->getInitializeAronTypeBlock(accessor_iterator);
                    b = CppBlock::mergeBlocks(b, b2);

                    b->addLine(accessor_iterator_container + ".push_back(" + accessor_iterator + ");");
                }
                b->addLine(getFullAronTypePtrTypename() + " " + accessor + " = " + getFullAronTypePtrTypename() + "(new " + getFullAronTypeTypename() + "(" + accessor_iterator_container + "));");
                return b;
            }

            CppBlockPtr AronTupleTypeNavigator::getResetBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());

                unsigned int i = 0;
                for (const auto& child : acceptedTypeNavigators)
                {
                    CppBlockPtr b2 = child->getResetBlock("std::get<" + std::to_string(i++) + ">(" + accessor + ")");
                    b = CppBlock::mergeBlocks(b, b2);
                }
                return b;
            }

            CppBlockPtr AronTupleTypeNavigator::getWriteBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("w.writeStartList();");

                unsigned int i = 0;
                for (const auto& type : acceptedTypeNavigators)
                {
                    CppBlockPtr b2 = type->getWriteBlock("std::get<" + std::to_string(i++) + ">(" + accessor + ")");
                    b = CppBlock::mergeBlocks(b, b2);
                }
                b->addLine("w.writeEndList();");
                return b;
            }

            CppBlockPtr AronTupleTypeNavigator::getReadBlock(const std::string& accessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("r.readStartList();");

                unsigned int i = 0;
                for (const auto& type : acceptedTypeNavigators)
                {
                    CppBlockPtr b2 = type->getReadBlock("std::get<" + std::to_string(i++) + ">(" + accessor + ")");
                    b = CppBlock::mergeBlocks(b, b2);
                }
                b->addLine("r.readEndList();");
                return b;
            }

            CppBlockPtr AronTupleTypeNavigator::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
            {
                CppBlockPtr b = CppBlockPtr(new CppBlock());
                b->addLine("if (not (" + accessor + " == " + otherInstanceAccessor + "))");
                b->addLine("\t return false;");
                return b;
            }
        }
    }
}

